import importlib.util as iutil
import socket
import sqlite3
from datetime import datetime
from os import path
from pathlib import Path
from typing import List, Dict
from uuid import uuid4

import numpy as np
import pandas as pd
from numpy import nan_to_num

from aequilibrae import Parameters
from aequilibrae.context import get_active_project
from aequilibrae.matrix import AequilibraeData
from aequilibrae.matrix import AequilibraeMatrix
from aequilibrae.paths.linear_approximation import LinearApproximation
from aequilibrae.paths.traffic_class import TrafficClass
from aequilibrae.paths.vdf import VDF, all_vdf_functions

spec = iutil.find_spec("openmatrix")
has_omx = spec is not None


class TrafficAssignment(object):
    """Traffic assignment class.

    For a comprehensive example on use, see the Use examples page.

    .. code-block:: python

        >>> from aequilibrae import Project
        >>> from aequilibrae.matrix import AequilibraeMatrix
        >>> from aequilibrae.paths import TrafficAssignment, TrafficClass

        >>> project = Project.from_path("/tmp/test_project")
        >>> project.network.build_graphs()

        >>> graph = project.network.graphs['c'] # we grab the graph for cars
        >>> graph.set_graph('free_flow_time') # let's say we want to minimize time
        >>> graph.set_skimming(['free_flow_time', 'distance']) # And will skim time and distance
        >>> graph.set_blocked_centroid_flows(True)

        >>> proj_matrices = project.matrices

        >>> demand = AequilibraeMatrix()
        >>> demand = proj_matrices.get_matrix("demand_omx")

        # We will only assign one user class stored as 'matrix' inside the OMX file
        >>> demand.computational_view(['matrix'])

        # Creates the assignment class
        >>> assigclass = TrafficClass("car", graph, demand)

        >>> assig = TrafficAssignment()

        # The first thing to do is to add at list of traffic classes to be assigned
        >>> assig.set_classes([assigclass])

        # Then we set the volume delay function
        >>> assig.set_vdf("BPR")  # This is not case-sensitive

        # And its parameters
        >>> assig.set_vdf_parameters({"alpha": "b", "beta": "power"})

        # The capacity and free flow travel times as they exist in the graph
        >>> assig.set_capacity_field("capacity")
        >>> assig.set_time_field("free_flow_time")

        # And the algorithm we want to use to assign
        >>> assig.set_algorithm('bfw')

        # Since we haven't checked the parameters file, let's make sure convergence criteria is good
        >>> assig.max_iter = 1000
        >>> assig.rgap_target = 0.00001

        >>> assig.execute() # we then execute the assignment

        # If you want, it is possible to access the convergence report
        >>> import pandas as pd
        >>> convergence_report = pd.DataFrame(assig.assignment.convergence_report)

        # Assignment results can be viewed as a Pandas DataFrame
        >>> results_df = assig.results()

        # Information on the assignment setup can be recovered with
        >>> info = assig.info()

        # Or save it directly to the results database
        >>> results = assig.save_results(table_name='base_year_assignment')

        # skims are here
        >>> avg_skims = assigclass.results.skims # blended ones
        >>> last_skims = assigclass._aon_results.skims # those for the last iteration
    """

    bpr_parameters = ["alpha", "beta"]
    all_algorithms = ["all-or-nothing", "msa", "frank-wolfe", "fw", "cfw", "bfw"]

    def __init__(self, project=None) -> None:
        """"""

        proj = project or get_active_project(must_exist=False)
        par = proj.parameters if proj else Parameters().parameters
        parameters = par["assignment"]["equilibrium"]

        self.__dict__["rgap_target"] = parameters["rgap"]
        self.__dict__["max_iter"] = parameters["maximum_iterations"]
        self.__dict__["vdf"] = VDF()
        self.__dict__["classes"] = []  # type: List[TrafficClass]
        self.__dict__["algorithm"] = None  # type: str
        self.__dict__["vdf_parameters"] = None  # type: list
        self.__dict__["time_field"] = None  # type: str
        self.__dict__["capacity_field"] = None  # type: str
        self.__dict__["assignment"] = None  # type: LinearApproximation
        self.__dict__["capacity"] = None  # type: np.ndarray
        self.__dict__["free_flow_tt"] = None  # type: np.ndarray
        self.__dict__["total_flow"] = None  # type: np.ndarray
        self.__dict__["congested_time"] = None  # type: np.ndarray
        self.__dict__["cores"] = None  # type: int
        self.__dict__["save_path_files"] = False  # type: bool

        self.__dict__["procedure_id"] = uuid4().hex
        self.__dict__["description"] = ""
        self.__dict__["procedure_date"] = str(datetime.today())
        self.__dict__["steps_below_needed_to_terminate"] = 1
        self.__dict__["project"] = proj

    def __setattr__(self, instance, value) -> None:
        check, value, message = self.__check_attributes(instance, value)
        if check:
            self.__dict__[instance] = value
        else:
            raise ValueError(message)

    def __check_attributes(self, instance, value):
        if instance == "rgap_target":
            if not isinstance(value, float):
                return False, value, "Relative gap needs to be a float"
            if isinstance(self.assignment, LinearApproximation):
                self.assignment.rgap_target = value
        elif instance == "max_iter":
            if not isinstance(value, int):
                return False, value, "Number of iterations needs to be an integer"
            if isinstance(self.assignment, LinearApproximation):
                self.assignment.max_iter = value
        elif instance == "vdf":
            v = value.lower()
            if v not in all_vdf_functions:
                return False, value, f"Volume-delay function {value} is not available"
            value = VDF()
            value.function = v
        elif instance == "classes":
            if isinstance(value, TrafficClass):
                value = [value]
            elif isinstance(value, list):
                for v in value:
                    if not isinstance(v, TrafficClass):
                        return False, value, "Traffic classes need to be proper AssignmentClass objects"
            else:
                raise ValueError("Traffic classes need to be proper AssignmentClass objects")
        elif instance == "vdf_parameters":
            if not self.__validate_parameters(value):
                return False, value, f"Parameter set is not valid: {value} "
        elif instance in ["time_field", "capacity_field"]:
            if not isinstance(value, str):
                return False, value, f"Value for {instance} is not string"
        elif instance == "cores":
            if not isinstance(value, int):
                return False, value, f"Value for {instance} is not integer"
        elif instance == "save_path_files":
            if not isinstance(value, bool):
                return False, value, f"Value for {instance} is not boolean"
        if instance not in self.__dict__:
            return False, value, f"trafficAssignment class does not have property {instance}"
        return True, value, ""

    def set_vdf(self, vdf_function: str) -> None:
        """
        Sets the Volume-delay function to be used

        :Arguments:
            **vdf_function** (:obj:`str`:) Name of the VDF to be used
        """
        self.vdf = vdf_function

    def set_classes(self, classes: List[TrafficClass]) -> None:
        """
        Sets Traffic classes to be assigned

        :Arguments:
            **classes** (:obj:`List[TrafficClass]`:) List of Traffic classes for assignment
        """

        ids = set([x.__id__ for x in classes])
        if len(ids) < len(classes):
            raise Exception("Classes need to be unique. Your list of classes has repeated items/IDs")
        self.classes = classes  # type: List[TrafficClass]

    def add_class(self, traffic_class: TrafficClass) -> None:
        """
        Adds a traffic class to the assignment

        :Arguments:
            **traffic_class** (:obj:`TrafficClass`:) Traffic class
        """

        ids = [x.__id__ for x in self.classes if x.__id__ == traffic_class.__id__]
        if len(ids) > 0:
            raise Exception("Traffic class already in the assignment")

        self.classes.append(traffic_class)

    def algorithms_available(self) -> list:
        """
        Returns all algorithms available for use

        :Returns:
            :obj:`list`: List of string values to be used with **set_algorithm**
        """
        return self.all_algorithms

    # TODO: Create procedure to check that travel times, capacities and vdf parameters are equal across all graphs
    # TODO: We also need procedures to check that all graphs are compatible (i.e. originated from the same network)
    def set_algorithm(self, algorithm: str):
        """
        Chooses the assignment algorithm. e.g. 'frank-wolfe', 'bfw', 'msa'

        'fw' is also accepted as an alternative to 'frank-wolfe'

        :Arguments:
            **algorithm** (:obj:`list`): Algorithm to be used
        """

        # First we instantiate the arrays we will be using over and over

        algo_dict = {i: i for i in self.all_algorithms}
        algo_dict["fw"] = "frank-wolfe"
        algo = algo_dict.get(algorithm.lower())

        if algo is None:
            raise AttributeError(f"Assignment algorithm not available. Choose from: {','.join(self.all_algorithms)}")

        if algo in ["all-or-nothing", "msa", "frank-wolfe", "cfw", "bfw"]:
            self.assignment = LinearApproximation(self, algo, project=self.project)
        else:
            raise Exception("Algorithm not listed in the case selection")

        self.__dict__["algorithm"] = algo

    def set_vdf_parameters(self, par: dict) -> None:
        """
        Sets the parameters for the Volume-delay function.

        Parameter values can be scalars (same values for the entire network) or network field names
        (link-specific values) - Examples: {'alpha': 0.15, 'beta': 4.0} or  {'alpha': 'alpha', 'beta': 'beta'}

        :Arguments:
            **par** (:obj:`dict`): Dictionary with all parameters for the chosen VDF

        """
        if self.classes is None or self.vdf.function.lower() not in all_vdf_functions:
            raise Exception("Before setting vdf parameters, you need to set traffic classes and choose a VDF function")
        self.__dict__["vdf_parameters"] = par
        pars = []
        if self.vdf.function in ["BPR", "BPR2", "CONICAL", "INRETS"]:
            for p1 in ["alpha", "beta"]:
                if p1 not in par:
                    raise ValueError(f"{p1} should exist in the set of parameters provided")
                p = par[p1]
                if isinstance(self.vdf_parameters[p1], str):
                    c = self.classes[0]
                    array = np.zeros(c.graph.graph.shape[0], c.graph.default_types("float"))
                    array[c.graph.graph.__supernet_id__] = c.graph.graph[p]
                else:
                    array = np.zeros(self.classes[0].graph.graph.shape[0], np.float64)
                    array.fill(self.vdf_parameters[p1])
                pars.append(array)

                if np.any(np.isnan(array)):
                    raise ValueError(f"At least one {p1} is NaN")

                if p1 == "alpha":
                    if array.min() < 0:
                        raise ValueError(f"At least one {p1} is smaller than zero")
                else:
                    if array.min() < 1:
                        raise ValueError(f"At least one {p1} is smaller than one. Results will make no sense")

        self.__dict__["vdf_parameters"] = pars

    def set_cores(self, cores: int) -> None:
        """Allows one to set the number of cores to be used AFTER traffic classes have been added

            Inherited from :obj:`AssignmentResults`

        :Arguments:
            **cores** (:obj:`int`): Number of CPU cores to use
        """
        if not self.classes:
            raise Exception("You need load traffic classes before overwriting the number of cores")

        self.cores = cores
        for c in self.classes:
            c.results.set_cores(cores)
            c._aon_results.set_cores(cores)

    def set_save_path_files(self, save_it: bool) -> None:
        """Turn path saving on or off.

        :Arguments:
            **save_it** (:obj:`bool`): Boolean to indicate whether paths should be saved
        """
        if self.classes is None:
            raise Exception("You need to set traffic classes before turning path saving on or off")

        # self.save_path_files = save_it
        for c in self.classes:
            c._aon_results.save_path_file = save_it

    def set_path_file_format(self, file_format: str) -> None:
        """Specify path saving format. Either parquet or feather.

        :Arguments:
            **file_format** (:obj:`str`): Name of file format to use for path files
        """
        if self.classes is None:
            raise Exception("You need to set traffic classes before specifying path saving options")

        if file_format == "feather":
            for c in self.classes:
                c._aon_results.write_feather = True
        elif file_format == "parquet":
            for c in self.classes:
                c._aon_results.write_feather = False
        else:
            raise Exception(f"Unsupported path file format {file_format} - only feather or parquet available.")

    def set_time_field(self, time_field: str) -> None:
        """
        Sets the graph field that contains free flow travel time -> e.g. 'fftime'

        :Arguments:
            **time_field** (:obj:`str`): Field name
        """

        if not self.classes:
            raise ValueError("Your need add at least one traffic classes first")

        c = self.classes[0]
        if time_field not in c.graph.graph.columns:
            raise ValueError("Field not in graph")

        if np.any(np.isnan(c.graph.graph[time_field].values)):
            raise ValueError("At least one link free-flow time is NaN")

        if c.graph.graph[time_field].values.min() <= 0:
            raise ValueError("There is at least one link with zero or negative free-flow time")

        self.__dict__["free_flow_tt"] = np.zeros(c.graph.graph.shape[0], c.graph.default_types("float"))
        self.__dict__["free_flow_tt"][c.graph.graph.__supernet_id__] = c.graph.graph[time_field]
        self.__dict__["congested_time"] = np.array(self.free_flow_tt, copy=True)
        self.__dict__["total_flow"] = np.zeros(self.free_flow_tt.shape[0], np.float64)
        self.time_field = time_field

    def set_capacity_field(self, capacity_field: str) -> None:
        """
        Sets the graph field that contains link capacity for the assignment period -> e.g. 'capacity1h'

        :Arguments:
            **capacity_field** (:obj:`str`): Field name
        """

        if not self.classes:
            raise ValueError("Your need add at least one traffic classes first")

        c = self.classes[0]
        if capacity_field not in c.graph.graph.columns:
            raise ValueError("Field not in graph")

        if np.any(np.isnan(c.graph.graph[capacity_field].values)):
            raise ValueError("At least one link capacity is NaN")

        if c.graph.graph[capacity_field].values.min() <= 0:
            raise ValueError("There is at least one link with zero or negative capacity")

        self.__dict__["cores"] = c.results.cores
        self.__dict__["capacity"] = np.zeros(c.graph.graph.shape[0], c.graph.default_types("float"))
        self.__dict__["capacity"][c.graph.graph.__supernet_id__] = c.graph.graph[capacity_field]
        self.capacity_field = capacity_field

    # TODO: This function actually needs to return a human-readable dictionary, and not one with
    #       tons of classes. Feeds into the class above
    # def load_assignment_spec(self, specs: dict) -> None:
    #     pass
    # def get_spec(self) -> dict:
    #     """Gets the entire specification of the assignment"""
    #     return deepcopy(self.__dict__)

    def __validate_parameters(self, kwargs) -> bool:
        if self.vdf == "":
            raise ValueError("First you need to set the Volume-Delay Function to use")

        par = list(kwargs.keys())
        q = [x for x in par if x not in self.bpr_parameters] + [x for x in self.bpr_parameters if x not in par]
        if len(q) > 0:
            raise ValueError("List of functions {} for vdf {} has an inadequate set of parameters".format(q, self.vdf))
        return True

    def execute(self) -> None:
        """Processes assignment"""
        self.assignment.execute()

    def save_results(self, table_name: str, keep_zero_flows=True, project=None) -> None:
        """Saves the assignment results to results_database.sqlite

        Method fails if table exists

        :Arguments:
            **table_name** (:obj:`str`): Name of the table to hold this assignment result
            **keep_zero_flows** (:obj:`bool`): Whether we should keep records for zero flows. Defaults to True
            **project** (:obj:`Project`, Optional): Project we want to save the results to. Defaults to the active project
        """

        df = self.results()
        if not keep_zero_flows:
            df = df[df.PCE_tot > 0]

        if not project:
            project = self.project or get_active_project()
        conn = sqlite3.connect(path.join(project.project_base_path, "results_database.sqlite"))
        df.to_sql(table_name, conn)
        conn.close()

        conn = project.connect()
        report = {"convergence": str(self.assignment.convergence_report), "setup": str(self.info())}
        data = [table_name, "traffic assignment", self.procedure_id, str(report), self.procedure_date, self.description]
        conn.execute(
            """Insert into results(table_name, procedure, procedure_id, procedure_report, timestamp,
                                            description) Values(?,?,?,?,?,?)""",
            data,
        )
        conn.commit()
        conn.close()

    def results(self) -> pd.DataFrame:
        """Prepares the assignment results as a Pandas DataFrame

        :Returns:
            **DataFrame** (:obj:`pd.DataFrame`): Pandas dataframe with all the assignment results indexed on link_id
        """

        idx = self.classes[0].graph.graph.__supernet_id__
        assig_results = [cls.results.get_load_results() for cls in self.classes]

        class1 = self.classes[0]
        res1 = assig_results[0]

        tot_flow = self.assignment.fw_total_flow[idx]
        voc = tot_flow / self.capacity[idx]
        congested_time = self.congested_time[idx]
        free_flow_tt = self.free_flow_tt[idx]

        fields = [
            "Congested_Time_AB",
            "Congested_Time_BA",
            "Congested_Time_Max",
            "Delay_factor_AB",
            "Delay_factor_BA",
            "Delay_factor_Max",
            "VOC_AB",
            "VOC_BA",
            "VOC_max",
            "PCE_AB",
            "PCE_BA",
            "PCE_tot",
        ]

        agg = AequilibraeData.empty(
            memory_mode=True,
            entries=res1.data.shape[0],
            field_names=fields,
            data_types=[np.float64] * len(fields),
            fill=np.nan,
            index=res1.data.index[:],
        )

        # Use the first class to get a graph -> network link ID mapping
        m = class1.results.get_graph_to_network_mapping()
        graph_ab, graph_ba = m.graph_ab_idx, m.graph_ba_idx
        agg.data["Congested_Time_AB"][m.network_ab_idx] = nan_to_num(congested_time[m.graph_ab_idx])
        agg.data["Congested_Time_BA"][m.network_ba_idx] = nan_to_num(congested_time[m.graph_ba_idx])
        agg.data["Congested_Time_Max"][:] = np.nanmax([agg.data.Congested_Time_AB, agg.data.Congested_Time_BA], axis=0)

        agg.data["Delay_factor_AB"][m.network_ab_idx] = nan_to_num(congested_time[graph_ab] / free_flow_tt[graph_ab])
        agg.data["Delay_factor_BA"][m.network_ba_idx] = nan_to_num(congested_time[graph_ba] / free_flow_tt[graph_ba])
        agg.data["Delay_factor_Max"][:] = np.nanmax([agg.data.Delay_factor_AB, agg.data.Delay_factor_BA], axis=0)

        agg.data["VOC_AB"][m.network_ab_idx] = nan_to_num(voc[m.graph_ab_idx])
        agg.data["VOC_BA"][m.network_ba_idx] = nan_to_num(voc[m.graph_ba_idx])
        agg.data["VOC_max"][:] = np.nanmax([agg.data.VOC_AB, agg.data.VOC_BA], axis=0)

        agg.data["PCE_AB"][m.network_ab_idx] = nan_to_num(tot_flow[m.graph_ab_idx])
        agg.data["PCE_BA"][m.network_ba_idx] = nan_to_num(tot_flow[m.graph_ba_idx])
        agg.data["PCE_tot"][:] = np.nansum([agg.data.PCE_AB, agg.data.PCE_BA], axis=0)

        assig_results.append(agg)

        dfs = [pd.DataFrame(aed.data) for aed in assig_results]
        dfs = [df.rename(columns={"index": "link_id"}).set_index("link_id") for df in dfs]
        df = pd.concat(dfs, axis=1)

        return df

    def report(self) -> pd.DataFrame:
        """Returns the assignment convergence report

        :Returns:
           **DataFrame** (:obj:`pd.DataFrame`): Convergence report
        """
        return pd.DataFrame(self.assignment.convergence_report)

    def info(self) -> dict:
        """Returns information for the traffic assignment procedure

        Dictionary contains keys  'Algorithm', 'Classes', 'Computer name', 'Procedure ID',
        'Maximum iterations' and 'Target RGap'.

        The classes key is also a dictionary with all the user classes per traffic class and their respective
        matrix totals

        :Returns:
            **info** (:obj:`dict`): Pandas dataframe with all the assignment results indexed on link_id
        """

        classes = {}

        for cls in self.classes:
            uclass = {}

            if len(cls.matrix.view_names) == 1:
                uclass["matrix_totals"] = {nm: np.sum(cls.matrix.matrix_view[:, :]) for nm in cls.matrix.view_names}
            else:
                uclass["matrix_totals"] = {
                    nm: np.sum(cls.matrix.matrix_view[:, :, i]) for i, nm in enumerate(cls.matrix.view_names)
                }
            uclass["network mode"] = cls.graph.mode
            uclass["Value-of-time"] = cls.vot
            uclass["PCE"] = cls.pce
            if cls.fixed_cost_field:
                uclass["Fixed cost field"] = cls.fixed_cost_field
                uclass["Fixed cost multiplier"] = cls.fc_multiplier
            uclass["save_path_files"] = cls._aon_results.save_path_file
            uclass["path_file_feather_format"] = cls._aon_results.write_feather

            classes[cls.__id__] = uclass

        info = {
            "Algorithm": self.algorithm,
            "Classes": classes,
            "Computer name": socket.gethostname(),
            "Maximum iterations": self.assignment.max_iter,
            "Procedure ID": self.procedure_id,
            "Target RGap": self.assignment.rgap_target,
        }
        return info

    def save_skims(self, matrix_name: str, which_ones="final", format="omx", project=None) -> None:
        """Saves the skims (if any) to the skim folder and registers in the matrix list

        :Arguments:
            **name** (:obj:`str`): Name of the matrix record to hold this matrix (same name used for file name)
            **which_ones** (:obj:`str`,optional): {'final': Results of the final iteration, 'blended': Averaged results for
            all iterations, 'all': Saves skims for both the final iteration and the blended ones} Default is 'final'
            **format** (:obj:`str`, `Optional`): File format ('aem' or 'omx'). Default is 'omx'
            **project** (:obj:`Project`, Optional): Project we want to save the results to. Defaults to the active project
        """
        mat_format = format.lower()
        if mat_format not in ["omx", "aem"]:
            raise ValueError("Matrix needs to be either OMX or native AequilibraE")
        if mat_format == "omx" and not has_omx:
            raise ImportError("OpenMatrix is not available on your system")

        if not project:
            project = self.project or get_active_project()

        mats = project.matrices

        for cls in self.classes:
            file_name = f"{matrix_name}_{cls.__id__}.{mat_format}"

            export_name = path.join(mats.fldr, file_name)

            if path.isfile(export_name):
                raise FileExistsError(f"{file_name} already exists. Choose a different name or matrix format")

            if mats.check_exists(matrix_name):
                raise FileExistsError(f"{matrix_name} already exists. Choose a different name")

            avg_skims = cls.results.skims  # type: AequilibraeMatrix

            # The ones for the last iteration are here
            last_skims = cls._aon_results.skims  # type: AequilibraeMatrix

            names = []
            if which_ones in ["final", "all"]:
                for core in last_skims.names:
                    names.append(f"{core}_final")

            if which_ones in ["blended", "all"]:
                for core in avg_skims.names:
                    names.append(f"{core}_blended")

            if not names:
                continue
            # Assembling a single final skim file can be done like this
            # We will want only the time for the last iteration and the distance averaged out for all iterations
            working_name = export_name if mat_format == "aem" else AequilibraeMatrix().random_name()

            kwargs = {
                "file_name": working_name,
                "zones": self.classes[0].graph.centroids.shape[0],
                "matrix_names": names,
                "memory_only": False,
            }

            # Create the matrix to manipulate
            out_skims = AequilibraeMatrix()
            out_skims.create_empty(**kwargs)

            out_skims.index[:] = self.classes[0].graph.centroids[:]
            out_skims.description = f"Assignment skim from procedure ID {self.procedure_id}. Class name {cls.__id__}"

            if which_ones in ["final", "all"]:
                for core in last_skims.names:
                    out_skims.matrix[f"{core}_final"][:, :] = last_skims.matrix[core][:, :]

            if which_ones in ["blended", "all"]:
                for core in avg_skims.names:
                    out_skims.matrix[f"{core}_blended"][:, :] = avg_skims.matrix[core][:, :]

            out_skims.matrices.flush()  # Make sure that all data went to the disk

            # If it were supposed to be an OMX, we export to one
            if mat_format == "omx":
                out_skims.export(export_name)

            out_skims.description = f"Skimming for assignment procedure. Class {cls.__id__}"
            # Now we create the appropriate record

            record = mats.new_record(f"{matrix_name}_{cls.__id__}", file_name)
            record.procedure_id = self.procedure_id
            record.timestamp = self.procedure_date
            record.procedure = "Traffic Assignment"
            record.description = out_skims.description
            record.save()

    def select_link_flows(self) -> Dict[str, pd.DataFrame]:
        """
        Returns a dataframe of the select link flows for each class
        """
        sl_flows = None  # stores the df for each class
        for cls in self.classes:
            # Save OD_matrices
            if cls._selected_links is None:
                continue
            df = cls.results.get_sl_results()
            # Create Values table
            df = pd.DataFrame(df.data)
            # Remap the dataframe names to add the prefix classname for each class
            cls_cols = {x: cls.__id__ + "_" + x if (x != "index") else "link_id" for x in df.columns}
            df.rename(columns=cls_cols, inplace=True)
            df.set_index("link_id", inplace=True)
            if sl_flows is None:
                sl_flows = df
            else:
                sl_flows.join(df)
        # sl_flows = pd.concat(class_flows, axis=1)
        return sl_flows

    def save_select_link_flows(self, table_name: str, project=None) -> None:
        """
        Saves the select link link flows for all classes into the results database. Additionally, it exports
        the OD matrices into OMX format.

        :Arguments:
            **table_name** (:obj:`str`): Name of the table being inserted to. Note the traffic class
            **project** (:obj:`Project`, `Optional`): Project we want to save the results to.
            Defaults to the active project
        """

        if not project:
            project = self.project or get_active_project()
        df = self.select_link_flows()
        conn = sqlite3.connect(path.join(project.project_base_path, "results_database.sqlite"))
        df.to_sql(table_name, conn)
        conn.close()
        # Create description table

        self.description = f"Select link analysis from {self.procedure_id}"
        conn = project.connect()
        report = {}
        data = [
            table_name,
            "select link",
            self.procedure_id,
            str(report),
            self.procedure_date,
            self.description,
        ]
        conn.execute(
            """Insert into results(table_name, procedure, procedure_id, procedure_report, timestamp,
                                            description) Values(?,?,?,?,?,?)""",
            data,
        )
        conn.commit()
        conn.close()

    def save_select_link_matrices(self, file_name: str) -> None:
        """
        Saves the Select Link matrices for each TrafficClass in the current TrafficAssignment class
        """

        for cls in self.classes:
            # Save OD_matrices
            if cls._selected_links is None:
                continue
            cls.results.select_link_od.export(str(Path(file_name).with_suffix(".omx")))

    def save_select_link_results(self, name: str) -> None:
        """
        Saves both the Select Link matrices and flow results at the same time, using the same name.

        .. note::
            Note the Select Link matrices will have _SL_matrices.omx appended to the end for ease of identification.
            e.g. save_select_link_results("Car") will result in the following names for the flows and matrices:
            Select Link Flows: inserts the select link flows for each class into the database with the table name:
            Car
            Select Link Matrices (only exports to OMX format):
            Car.omx

        :Arguments:
            **name** (:obj:`str`): name of the matrices
        """
        self.save_select_link_flows(name)
        self.save_select_link_matrices(name)



import numpy as np

from aequilibrae import global_logger
from aequilibrae.paths.graph import Graph

try:
    from aequilibrae.paths.AoN import update_path_trace, path_computation
except ImportError as ie:
    global_logger.warning(f"Could not import procedures from the binary. {ie.args}")


class PathResults:
    """Path computation result holder

    .. code-block:: python

          >>> from aequilibrae import Project
          >>> from aequilibrae.paths.results import PathResults

          >>> proj = Project.from_path("/tmp/test_project")
          >>> proj.network.build_graphs()

          # Mode c is car in this project
          >>> car_graph = proj.network.graphs['c']

          # minimize distance
          >>> car_graph.set_graph('distance')

          # If you want to compute skims
          # It does increase path computation time substantially
          >>> car_graph.set_skimming(['distance', 'free_flow_time'])

          >>> res = PathResults()
          >>> res.prepare(car_graph)
          >>> res.compute_path(1, 17)

          # res.milepost contains the milepost corresponding to each node along the path
          # res.path_nodes contains the sequence of nodes that form the path
          # res.path  contains the sequence of links that form the path
          # res.path_link_directions contains the link directions corresponding to the above links
          # res.skims contain all skims requested when preparing the graph

          # Update all the outputs mentioned above for destination 9. Same origin: 1
          >>> res.update_trace(9)

          # clears all computation results
          >>> res.reset()
    """

    def __init__(self) -> None:
        self.predecessors = None
        self.connectors = None
        self.skims = None
        self._skimming_array = None
        self.path = None
        self.path_nodes = None
        self.path_link_directions = None
        self.milepost = None
        self.reached_first = None
        self.origin = None
        self.destination = None
        self.graph: Graph = None
        self.links = -1
        self.nodes = -1
        self.zones = -1
        self.num_skims = -1
        self.__integer_type = None
        self.__float_type = None
        self.__graph_id__ = None
        self.__graph_sum = None

    def compute_path(self, origin: int, destination: int) -> None:
        """
        Computes the path between two nodes in the network

        :Arguments:
            **origin** (:obj:`int`): Origin for the path

            **destination** (:obj:`int`): Destination for the path
        """

        if self.graph is None:
            raise Exception("You need to set graph skimming before you compute a path")

        path_computation(origin, destination, self.graph, self)
        if self.graph.skim_fields:
            self.skims.fill(np.inf)
            self.skims[self.graph.all_nodes, :] = self._skimming_array[:-1, :]
            self.skims[self.skims > self.__graph_sum] = np.inf

    def prepare(self, graph: Graph) -> None:
        """
        Prepares the object with dimensions corresponding to the graph object

        :Arguments:
            **graph** (:obj:`Graph`): Needs to have been set with number of centroids and list of skims (if any)
        """

        if not graph.cost_field:
            raise Exception('Cost field needs to be set for cost computation. use graph.set_graph("your_cost_field")')

        self.__integer_type = graph.default_types("int")
        self.__float_type = graph.default_types("float")
        self.nodes = graph.num_nodes + 1
        self.zones = graph.centroids + 1
        self.links = graph.num_links + 1
        self.num_skims = len(graph.skim_fields)

        self.predecessors = np.zeros(self.nodes, dtype=self.__integer_type)
        self.connectors = np.zeros(self.nodes, dtype=self.__integer_type)
        self.reached_first = np.zeros(self.nodes, dtype=self.__integer_type)
        if self.num_skims:
            self.skims = np.empty((np.max(graph.all_nodes) + 1, self.num_skims), self.__float_type)
            self.skims.fill(np.inf)
            self._skimming_array = np.zeros((self.nodes, self.num_skims), self.__float_type)
        else:
            self._skimming_array = np.zeros((1, 2), self.__float_type)

        self.__graph_id__ = graph.__id__
        # We can imagine somebody creating a worst-case scenario network (imagining that turn penalties are considered)
        # where one needs to traverse all links (or almost all) in both directions.
        self.__graph_sum = 2 * graph.cost.sum()
        self.graph = graph

    def reset(self) -> None:
        """
        Resets object to prepared and pre-computation state
        """
        if self.predecessors is not None:
            self.predecessors.fill(-1)
            self.connectors.fill(-1)
            if self.skims is not None:
                self.skims.fill(np.inf)
            self._skimming_array.fill(np.inf)
            self.path = None
            self.path_nodes = None
            self.path_link_directions = None
            self.milepost = None

        else:
            raise ValueError("Exception: Path results object was not yet prepared/initialized")

    def update_trace(self, destination: int) -> None:
        """
        Updates the path's nodes, links, skims and mileposts

        It does not re-compute the path tree, so it saves most of the computation time

        :Arguments:
            **destination** (:obj:`int`): ID of the node we are computing the path too
        """
        if not isinstance(destination, int):
            raise TypeError("destination needs to be an integer")

        if destination >= self.graph.nodes_to_indices.shape[0]:
            raise ValueError("destination out of the range of node numbers in the graph")

        update_path_trace(self, destination, self.graph)



from math import pi, sqrt
import numpy as np
from scipy.cluster.vq import kmeans2, whiten
from scipy.spatial.distance import cdist
import shapely.wkb
from shapely.geometry import LineString

INFINITE_CAPACITY = 99999


def connector_creation(geo, zone_id: int, srid: int, mode_id: str, network, link_types="", connectors=1):
    if len(mode_id) > 1:
        raise Exception("We can only add centroid connectors for one mode at a time")

    conn = network.project.connect()
    curr = conn.cursor()
    logger = network.project.logger
    curr.execute("select count(*) from nodes where node_id=?", [zone_id])
    if curr.fetchone() is None:
        logger.warning("This centroid does not exist. Please create it first")
        return

    proj_nodes = network.nodes
    node = proj_nodes.get(zone_id)
    curr.execute("select count(*) from links where a_node=? and instr(modes,?) > 0", [zone_id, mode_id])
    if curr.fetchone()[0] > 0:
        logger.warning("Mode is already connected")
        return

    if len(link_types) > 0:
        lt = f"*[{link_types}]*"
    else:
        curr.execute("Select link_type_id from link_types")
        lt = "".join([x[0] for x in curr.fetchall()])
        lt = f"*[{lt}]*"

    sql = """select node_id, ST_asBinary(geometry), modes, link_types from nodes where ST_Within(geometry, GeomFromWKB(?, ?)) and
                    (nodes.rowid in (select rowid from SpatialIndex where f_table_name = 'nodes' and
                    search_frame = GeomFromWKB(?, ?)))
            and link_types glob ? and instr(modes, ?)>0"""

    # We expand the area by its average radius until it is 20 times
    # beginning with a strict search within the zone
    buffer = 0
    increase = sqrt(geo.area / pi)
    dt = []
    while dt == [] and buffer <= increase * 10:
        wkb = geo.buffer(buffer).wkb
        curr.execute(sql, [wkb, srid, wkb, srid, lt, mode_id])
        dt = curr.fetchall()
        buffer += increase

    if buffer > increase:
        msg = f"Could not find node inside zone {zone_id}. Search area was expanded until we found a suitable node"
        logger.warning(msg)
    if dt == []:
        logger.warning(
            f"FAILED! Could not find suitable nodes to connect within 5 times the diameter of zone {zone_id}."
        )
        return

    coords = []
    nodes = []
    for node_id, wkb, modes, link_types in dt:
        geo = shapely.wkb.loads(wkb)
        coords.append([geo.x, geo.y])
        nodes.append(node_id)

    num_connectors = connectors
    if len(nodes) == 0:
        raise Exception("We could not find any candidate nodes that satisfied your criteria")
    elif len(nodes) < connectors:
        logger.warning(
            f"We have fewer possible nodes than required connectors for zone {zone_id}. Will connect all of them."
        )
        num_connectors = len(nodes)

    if num_connectors == len(coords):
        all_nodes = nodes
    else:
        features = np.array(coords)
        whitened = whiten(features)
        centroids, allocation = kmeans2(whitened, num_connectors)

        all_nodes = set()
        for i in range(num_connectors):
            nds = [x for x, y in zip(nodes, list(allocation)) if y == i]
            centr = centroids[i]
            positions = [x for x, y in zip(whitened, allocation) if y == i]
            if positions:
                dist = cdist(np.array([centr]), np.array(positions)).flatten()
                node_to_connect = nds[dist.argmin()]
                all_nodes.add(node_to_connect)

    nds = list(all_nodes)
    data = [zone_id] + nds
    curr.execute(f'select b_node from links where a_node=? and b_node in ({",".join(["?"] * len(nds))})', data)

    data = [x[0] for x in curr.fetchall()]

    if data:
        qry = ",".join(["?"] * len(data))
        dt = [mode_id, zone_id] + data
        curr.execute(f"Update links set modes=modes || ? where a_node=? and b_node in ({qry})", dt)
        nds = [x for x in nds if x not in data]
        logger.warning(f"Mode {mode_id} added to {len(data)} existing centroid connectors for zone {zone_id}")
        conn.commit()

    curr.close()
    links = network.links

    for node_to_connect in nds:
        link = links.new()
        node_to = proj_nodes.get(node_to_connect)
        link.geometry = LineString([node.geometry, node_to.geometry])
        link.modes = mode_id
        link.direction = 0
        link.link_type = "centroid_connector"
        link.name = f"centroid connector zone {zone_id}"
        link.capacity_ab = INFINITE_CAPACITY
        link.capacity_ba = INFINITE_CAPACITY
        link.save()
    if nds:
        logger.warning(f"{len(nds)} new centroid connectors for mode {mode_id} added for centroid {zone_id}")

    conn.commit()



import os
import tempfile
from unittest import TestCase

import numpy as np

from aequilibrae.distribution import SyntheticGravityModel, GravityApplication
from aequilibrae.matrix import AequilibraeData, AequilibraeMatrix
from aequilibrae.parameters import Parameters

zones = 10

# row vector
args = {"entries": zones, "field_names": ["rows"], "data_types": [np.float64], "memory_mode": True}

row_vector = AequilibraeData()
row_vector.create_empty(**args)
row_vector.index[:] = np.arange(row_vector.entries) + 100
row_vector.rows[:] = row_vector.index[:] + np.random.rand(zones)[:]

# column vector
args["field_names"] = ["columns"]
column_vector = AequilibraeData()
column_vector.create_empty(**args)
column_vector.index[:] = np.arange(column_vector.entries) + 100
column_vector.columns[:] = column_vector.index[:] + np.random.rand(zones)[:]

# balance vectors
column_vector.columns[:] = column_vector.columns[:] * (row_vector.rows.sum() / column_vector.columns.sum())

# Impedance matrix_procedures
name_test = os.path.join(tempfile.gettempdir(), "aequilibrae_matrix_test.aem")

args = {"file_name": name_test, "zones": zones, "matrix_names": ["impedance"]}

matrix = AequilibraeMatrix()
matrix.create_empty(**args)

# randoms = np.random.randint(5, size=(2, 4))
matrix.matrices[:, :, 0] = np.random.rand(zones, zones)[:, :]
matrix.index[:] = np.arange(matrix.zones) + 100
matrix.computational_view(["impedance"])

model_expo = SyntheticGravityModel()
model_expo.function = "EXPO"
model_expo.beta = 0.1

model_gamma = SyntheticGravityModel()
model_gamma.function = "GAMMA"
model_gamma.beta = 0.1
model_gamma.alpha = -0.2

model_power = SyntheticGravityModel()
model_power.function = "POWER"
model_power.alpha = -0.2


class TestGravityApplication(TestCase):
    def setUp(self):
        # GravityApplication requires an object that has a `parameters` attribute. `Parameters` fits
        # this requirement, so that we don't need to create a full project
        self.proj = Parameters()

    def test_apply(self):
        args = {
            "impedance": matrix,
            "rows": row_vector,
            "row_field": "rows",
            "columns": column_vector,
            "column_field": "columns",
        }

        models = [("EXPO", model_expo), ("POWER", model_power), ("GAMMA", model_gamma)]

        for model_name, model_obj in models:
            args["model"] = model_obj
            distributed_matrix = GravityApplication(project=self.proj, **args)
            distributed_matrix.apply()

            if distributed_matrix.gap > distributed_matrix.parameters["convergence level"]:
                self.fail(f"Gravity application did not converge for model {model_name}")



import sqlite3
from warnings import warn
from math import sqrt
from uuid import uuid4
from random import randint
from shutil import copytree, rmtree
from os.path import join
from tempfile import gettempdir
from unittest import TestCase
from shapely.geometry import Point, MultiPolygon, LineString, MultiLineString
import shapely.wkb
from aequilibrae import Project
from ...data import siouxfalls_project
from aequilibrae.utils.create_example import create_example


class TestZone(TestCase):
    def setUp(self) -> None:
        self.temp_proj_folder = join(gettempdir(), uuid4().hex)
        copytree(siouxfalls_project, self.temp_proj_folder)
        self.proj = Project()
        self.proj.open(self.temp_proj_folder)

    def tearDown(self) -> None:
        self.proj.close()
        try:
            rmtree(self.temp_proj_folder)
        except Exception as e:
            warn(f"Error: {e.args}")

    def test_delete(self):
        zones = self.proj.zoning
        zone_downtown = zones.get(3)
        zone_downtown.delete()

        with self.assertRaises(ValueError):
            _ = zones.get(3)

    def test_save(self):
        zones = self.proj.zoning
        zn = zones.get(2)
        area = randint(0, 9999999999)
        zn.area = area
        zn.save()

        curr = self.proj.conn.cursor()
        curr.execute("Select area from Zones where zone_id=2")
        self.assertEqual(curr.fetchone()[0], area, "Zone didn't save area properly")

        geo = Point(0, 0).buffer(1)
        zn.geometry = geo
        zn.save()
        curr = self.proj.conn.cursor()
        curr.execute("Select asBinary(geometry) from Zones where zone_id=2")
        wkb = curr.fetchone()[0]
        self.assertEqual(shapely.wkb.loads(wkb), MultiPolygon([geo]), "Zone didn't save geometry properly")

        zn2 = zones.get(1)
        geo = MultiPolygon([Point(0, 0).buffer(1)])
        zn2.geometry = geo
        zn2.save()
        curr = self.proj.conn.cursor()
        curr.execute("Select asBinary(geometry) from Zones where zone_id=1")
        wkb = curr.fetchone()[0]
        self.assertEqual(shapely.wkb.loads(wkb), geo, "Zone didn't save geometry properly")

    def __change_project(self):
        self.proj.close()
        self.proj = Project()
        self.proj = create_example(join(gettempdir(), uuid4().hex), "nauru")
        zones = 5
        network = self.proj.network
        nodes = network.nodes

        geo = network.convex_hull()

        zone_area = geo.area / zones
        zone_side = sqrt(2 * sqrt(3) * zone_area / 9)

        extent = network.extent()

        curr = self.proj.conn.cursor()
        b = extent.bounds
        curr.execute(
            "select st_asbinary(HexagonalGrid(GeomFromWKB(?), ?, 0, GeomFromWKB(?)))",
            [extent.wkb, zone_side, Point(b[2], b[3]).wkb],
        )
        grid = curr.fetchone()[0]
        grid = shapely.wkb.loads(grid)

        grid = [p for p in grid.geoms if p.intersects(geo)]

        zoning = self.proj.zoning
        for i, zone_geo in enumerate(grid):
            zone = zoning.new(i + 1)
            zone.geometry = zone_geo
            zone.save()

            node = nodes.get(i + 1)
            node.renumber(i + 10001)

    def test_add_centroid(self):
        self.__change_project()
        zones = self.proj.zoning
        nodes = self.proj.network.nodes
        network = self.proj.network
        zone1 = zones.get(1)
        tot = network.count_centroids()
        zone1.add_centroid(None)
        self.assertEqual(tot + 1, network.count_centroids(), "Added less than it should've")

        tot = network.count_centroids()
        zone1.add_centroid(None)
        zone1.add_centroid(Point(0, 0))
        self.assertEqual(tot, network.count_centroids(), "Added more than should've")
        node1 = nodes.get(1)
        self.assertEqual(node1.geometry, zone1.geometry.centroid)

        zone2 = zones.get(2)
        zone2.add_centroid(Point(0, 0))

        node2 = nodes.get(2)
        self.assertEqual(node2.geometry, Point(0, 0))

        # Tests ne behaviour to deal with centroids that would fall exactly on top of existing nodes
        point_that_should = zone1.geometry.centroid
        self.__change_project()
        zones = self.proj.zoning
        network = self.proj.network
        nd = network.nodes.get(1000)
        nd.geometry = point_that_should
        nd.save()

        zone1 = zones.get(1)
        with self.assertRaises(sqlite3.IntegrityError):
            zone1.add_centroid(None, robust=False)
        zone1.add_centroid(None, robust=True)

    def test_connect_mode(self):
        self.__change_project()
        curr = self.proj.conn.cursor()
        zones = self.proj.zoning

        zone1 = zones.get(1)
        zone1.add_centroid(None)

        zone1.connect_mode("c")

        curr.execute("Select count(*) from links where a_node=?", [1])
        self.assertIsNot(0, curr.fetchone()[0], "failed to add connectors")

        zone1.connect_mode("t")
        curr.execute("""Select count(*) from links where a_node=? and instr(modes,'t')>0""", [1])
        self.assertIsNot(0, curr.fetchone()[0], "failed to add connectors for mode t")

        # Cannot connect a centroid that does not exist
        with self.assertRaises(ValueError):
            zone2 = zones.get(2)
            zone2.connect_mode("c")

    def test_disconnect_mode(self):
        self.__change_project()
        curr = self.proj.conn.cursor()
        zones = self.proj.zoning
        zone1 = zones.get(1)
        zone1.add_centroid(None)

        zone1.connect_mode("c")
        zone1.connect_mode("w")
        curr.execute("""select COUNT(*) from links where a_node=1""")
        tot = curr.fetchone()[0]
        curr.execute("""Update links set modes = modes || 'w' where instr(modes,'w')=0""")
        self.proj.conn.commit()
        zone1.disconnect_mode("w")

        curr.execute("""select COUNT(*) from links where a_node=1""")
        self.assertIsNot(tot, curr.fetchone()[0], "failed to delete links")

        curr.execute("""Select count(*) from links where a_node=1 and instr(modes,'w')>0""")
        self.assertEqual(curr.fetchone()[0], 0, "Failed to remove mode from all connectors")

    def test_get_closest_zone(self):
        pt_in = Point(-96.7716, 43.6069)
        pt_out = Point(-96.7754, 43.5664)
        self.assertEqual(self.proj.zoning.get_closest_zone(pt_in), 1)
        self.assertEqual(self.proj.zoning.get_closest_zone(pt_out), 3)

        line_in = LineString([(-96.7209, 43.6132), (-96.7033, 43.61316)])
        line_out = LineString([(-96.7473, 43.6046), (-96.7341, 43.6046)])
        self.assertEqual(self.proj.zoning.get_closest_zone(line_in), 2)
        self.assertEqual(self.proj.zoning.get_closest_zone(line_out), 2)

        multi_line_in = MultiLineString(
            [((-96.7589, 43.5692), (-96.7531, 43.5807)), ((-96.7531, 43.5807), (-96.7504, 43.5704))]
        )
        multi_line_out = MultiLineString(
            [((-96.7716, 43.5769), (-96.7683, 43.5801)), ((-96.7683, 43.5801), (-96.7574, 43.5784))]
        )
        self.assertEqual(self.proj.zoning.get_closest_zone(multi_line_in), 3)
        self.assertEqual(self.proj.zoning.get_closest_zone(multi_line_out), 3)



import tempfile
from os.path import dirname, abspath, join

data_folder = join(dirname(dirname(abspath(__file__))), "data")

# For the graph tests
test_network = join(dirname(dirname(abspath(__file__))), "data", "Final_Network.shp")
test_graph = join(dirname(dirname(abspath(__file__))), "data", "test_graph.aeg")
path_test = tempfile.gettempdir()
triangle_graph_blocking = join(dirname(dirname(abspath(__file__))), "data", "blocking_triangle_graph_project")

gtfs_folder = join(dirname(dirname(abspath(__file__))), "data/gtfs")
gtfs_zip = join(dirname(dirname(abspath(__file__))), "data/gtfs.zip")
gtfs_db_output = join(path_test, "test.db")

# For the skimming test


# For the matrix test
omx_example = join(dirname(dirname(abspath(__file__))), "data/test_omx.omx")
no_index_omx = join(dirname(dirname(abspath(__file__))), "data/no_index.omx")

# For project tests
project_file = join(dirname(dirname(abspath(__file__))), "data", "AequilibraE_Project.sqlite")

# For Traffic Assignment tests
siouxfalls_project = join(dirname(dirname(abspath(__file__))), "data/SiouxFalls_project")
siouxfalls_demand = join(dirname(dirname(abspath(__file__))), "data/SiouxFalls_project/matrices", "SiouxFalls.omx")
siouxfalls_skims = join(dirname(dirname(abspath(__file__))), "data/SiouxFalls_project/matrices", "sfalls_skims.omx")

#
no_triggers_project = join(dirname(dirname(abspath(__file__))), "data/no_triggers_project")

st_varent_network = join(dirname(abspath(__file__)), "St_Varent_issue307.zip")



import importlib.util as iutil
import os
import platform
from os.path import dirname, join

import numpy as np
from Cython.Distutils import build_ext
from setuptools import Extension
from setuptools import setup, find_packages

with open("__version__.py") as f:
    exec(f.read())

include_dirs = [np.get_include()]
if iutil.find_spec("pyarrow") is not None:
    import pyarrow as pa

    include_dirs.append(pa.get_include())

is_win = "WINDOWS" in platform.platform().upper()
is_mac = any([e in platform.platform().upper() for e in ["MACOS", "DARWIN"]])
prefix = "/" if is_win else "-f"
cpp_std = "/std:c++17" if is_win else "-std=c++17"
compile_args = [cpp_std, f"{prefix}openmp"]
compile_args += ["-Wno-unreachable-code"] if is_mac else []
link_args = [f"{prefix}openmp"]

ext_mod_aon = Extension(
    "aequilibrae.paths.AoN",
    [join("aequilibrae", "paths", "AoN.pyx")],
    extra_compile_args=compile_args,
    extra_link_args=link_args,
    define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
    include_dirs=include_dirs,
    language="c++",
)

ext_mod_ipf = Extension(
    "aequilibrae.distribution.ipf_core",
    [join("aequilibrae", "distribution", "ipf_core.pyx")],
    extra_compile_args=compile_args,
    extra_link_args=link_args,
    define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
    include_dirs=include_dirs,
    language="c++",
)

with open("requirements.txt", "r") as fl:
    install_requirements = [x.strip() for x in fl.readlines()]

pkgs = [pkg for pkg in find_packages()]

pkg_data = {
    "aequilibrae.reference_files": ["spatialite.sqlite", "nauru.zip", "sioux_falls.zip", "coquimbo.zip"],
    "aequilibrae.paths": ["parameters.pxi", "*.pyx"],
    "aequilibrae.distribution": ["*.pyx"],
    "aequilibrae": ["./parameters.yml", "../requirements.txt"],
    "aequilibrae.project": [
        "database_specification/network/tables/*.*",
        "database_specification/network/triggers/*.*",
        "database_specification/transit/tables/*.*",
        "database_specification/transit/triggers/*.*",
    ],
}
loose_modules = ["__version__", "parameters"]

if __name__ == "__main__":
    setup(
        name="aequilibrae",
        version=release_version,  # noqa: F821
        install_requires=install_requirements,
        packages=pkgs,
        package_dir={"": "."},
        py_modules=loose_modules,
        package_data=pkg_data,
        zip_safe=False,
        description="A package for transportation modeling",
        author="Pedro Camargo",
        author_email="c@margo.co",
        url="https://github.com/AequilibraE/aequilibrae",
        license="See license.txt",
        classifiers=[
            "Programming Language :: Python",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
        ],
        cmdclass={"build_ext": build_ext},
        ext_modules=[ext_mod_aon, ext_mod_ipf],
    )



from unittest import TestCase
from aequilibrae.paths.results import AssignmentResults
import multiprocessing as mp


class TestAssignmentResults(TestCase):
    # def test_prepare(self):
    #     self.fail()
    #
    # def test_reset(self):
    #     self.fail()
    #
    # def test_total_flows(self):
    #     self.fail()

    def test_set_cores(self):
        a = AssignmentResults()
        a.set_cores(10)

        with self.assertRaises(ValueError):
            a.set_cores(1.3)

        a.set_cores(-2)
        self.assertEqual(a.cores, max(1, mp.cpu_count() - 2))

    def test_set_save_path_file(self):
        a = AssignmentResults()

        # Never save by default
        self.assertEqual(a.save_path_file, False)

    # def test_set_critical_links(self):
    #     self.fail()
    #
    # def test_get_load_results(self):
    #     self.fail()
    #
    # def test_save_to_disk(self):
    #     self.fail()



import datetime
import os
import random
from os.path import join
from shutil import copyfile
import openmatrix as omx
from unittest import TestCase
from tempfile import gettempdir
import numpy as np
import uuid
from aequilibrae.matrix import AequilibraeMatrix
from ...data import omx_example, no_index_omx, siouxfalls_skims
import pandas as pd
import pytest

zones = 50


class TestAequilibraeMatrix(TestCase):
    matrix = None

    def setUp(self) -> None:
        temp_folder = gettempdir()
        self.sf_skims = join(temp_folder, f"Aequilibrae_matrix_{uuid.uuid4()}.omx")
        copyfile(siouxfalls_skims, self.sf_skims)
        self.name_test = join(temp_folder, f"Aequilibrae_matrix_{uuid.uuid4()}.aem")
        self.copy_matrix_name = join(temp_folder, f"Aequilibrae_matrix_{uuid.uuid4()}.aem")
        self.csv_export_name = join(temp_folder, f"Aequilibrae_matrix_{uuid.uuid4()}.csv")
        self.omx_export_name = join(temp_folder, f"Aequilibrae_matrix_{uuid.uuid4()}.omx")

        if self.matrix is not None:
            return
        args = {
            "file_name": self.name_test,
            "zones": zones,
            "matrix_names": ["mat", "seed", "dist"],
            "index_names": ["my_indices"],
            "memory_only": False,
        }

        self.matrix = AequilibraeMatrix()
        self.matrix.create_empty(**args)

        self.matrix.index[:] = np.arange(self.matrix.zones) + 100
        self.matrix.matrices[:, :, 0] = np.random.rand(self.matrix.zones, self.matrix.zones)
        self.matrix.matrices[:, :, 0] = self.matrix.mat * (1000 / np.sum(self.matrix.mat))
        self.matrix.setName("Test matrix - " + str(random.randint(1, 10)))
        self.matrix.setDescription("Generated at " + datetime.datetime.now().strftime("%I:%M%p on %B %d, %Y"))
        self.new_matrix = self.matrix

    def tearDown(self) -> None:
        try:
            del self.matrix
            os.remove(self.name_test) if os.path.exists(self.name_test) else None
            os.remove(self.csv_export_name) if os.path.exists(self.csv_export_name) else None
            os.remove(self.copy_matrix_name) if os.path.exists(self.copy_matrix_name) else None
            os.remove(self.omx_export_name) if os.path.exists(self.omx_export_name) else None
        except Exception as e:
            print(f"Could not delete.  {e.args}")

    def test_load(self):
        self.new_matrix = AequilibraeMatrix()
        # Cannot load OMX file with no indices
        with self.assertRaises(LookupError):
            self.new_matrix.load(no_index_omx)

        self.new_matrix = AequilibraeMatrix()
        self.new_matrix.load(self.name_test)
        del self.new_matrix

    def test_computational_view(self):
        self.new_matrix.computational_view(["mat", "seed"])
        self.new_matrix.mat.fill(0)
        self.new_matrix.seed.fill(0)
        if self.new_matrix.matrix_view.shape[2] != 2:
            self.fail("Computational view returns the wrong number of matrices")

        self.new_matrix.computational_view(["mat"])
        self.new_matrix.matrix_view[:, :] = np.arange(zones**2).reshape(zones, zones)
        if np.sum(self.new_matrix.mat) != np.sum(self.new_matrix.matrix_view):
            self.fail("Assigning to matrix view did not work")
        self.new_matrix.setName("Test matrix - " + str(random.randint(1, 10)))
        self.new_matrix.setDescription("Generated at " + datetime.datetime.now().strftime("%I:%M%p on %B %d, %Y"))
        del self.new_matrix

    def test_computational_view_with_omx(self):
        self.new_matrix = AequilibraeMatrix()
        self.new_matrix.load(omx_example)

        arrays = ["m1", "m2"]
        self.new_matrix.computational_view(arrays)
        total_mats = np.sum(self.new_matrix.matrix_view)

        self.new_matrix.computational_view([arrays[0]])
        total_m1 = np.sum(self.new_matrix.matrix_view)

        self.new_matrix.close()

        omx_file = omx.open_file(omx_example, "r")

        m1 = np.array(omx_file["m1"]).sum()
        m2 = np.array(omx_file["m2"]).sum()

        self.assertEqual(m1 + m2, total_mats)
        self.assertEqual(m1, total_m1)

        omx_file.close()
        del omx_file

    def test_copy(self):
        # test in-memory matrix_procedures copy

        matrix_copy = self.new_matrix.copy(self.copy_matrix_name, cores=["mat"])

        if not np.array_equal(matrix_copy.mat, self.new_matrix.mat):
            self.fail("Matrix copy was not perfect")
        matrix_copy.close()
        del matrix_copy

    def test_copy_memory_only(self):
        # test in-memory matrix_procedures copy

        matrix_copy = self.new_matrix.copy(self.copy_matrix_name, cores=["mat"], memory_only=True)

        for orig, new_arr in [[matrix_copy.mat, self.new_matrix.mat], [matrix_copy.index, self.new_matrix.index]]:
            if not np.array_equal(orig, new_arr):
                self.fail("Matrix copy was not perfect")
        matrix_copy.close()

    def test_export_to_csv(self):
        self.new_matrix.export(self.csv_export_name)
        df = pd.read_csv(self.csv_export_name)
        df.fillna(0, inplace=True)
        self.assertEqual(df.shape[0], 2500, "Exported wrong size")
        self.assertEqual(df.shape[1], 5, "Exported wrong size")
        self.assertAlmostEqual(df.mat.sum(), np.nansum(self.new_matrix.matrices), 5, "Exported wrong matrix total")

    def test_export_to_omx(self):
        self.new_matrix.export(self.omx_export_name)

        omxfile = omx.open_file(self.omx_export_name, "r")

        # Check if matrices values are compatible
        for m in self.new_matrix.names:
            sm = np.nansum(self.new_matrix.matrix[m])
            sm2 = np.nansum(np.array(omxfile[m]))

            self.assertEqual(sm, sm2, "Matrix {} was exported with the wrong value".format(m))
        del omxfile

    def test_nan_to_num(self):
        m = self.new_matrix.mat.sum() - self.new_matrix.mat[1, 1]
        self.new_matrix.computational_view(["mat", "seed"])
        self.new_matrix.nan_to_num()
        self.new_matrix.matrices[1, 1, 0] = np.nan
        self.new_matrix.computational_view(["mat"])
        self.new_matrix.nan_to_num()

        if abs(m - self.new_matrix.mat.sum()) > 0.000000000001:
            self.fail("Total for mat matrix not maintained")
        del self.new_matrix

    def test_copy_from_omx(self):
        temp_file = AequilibraeMatrix().random_name()
        a = AequilibraeMatrix()
        a.create_from_omx(temp_file, omx_example)

        omxfile = omx.open_file(omx_example, "r")

        # Check if matrices values are compatible
        for m in ["m1", "m2", "m3"]:
            sm = a.matrix[m].sum()
            sm2 = np.array(omxfile[m]).sum()
            if sm != sm2:
                self.fail("Matrix {} was copied with the wrong value".format(m))

        if np.any(a.index[:] != np.array(list(omxfile.mapping("taz").keys()))):
            self.fail("Index was not created properly")
        a.close()
        del a
        del omxfile

    def test_copy_from_omx_long_name(self):
        temp_file = AequilibraeMatrix().random_name()
        a = AequilibraeMatrix()

        with self.assertRaises(ValueError):
            a.create_from_omx(temp_file, omx_example, robust=False)
        del a

    def test_copy_omx_wrong_content(self):
        # Check if we get a result if we try to copy non-existing cores
        temp_file = AequilibraeMatrix().random_name()
        a = AequilibraeMatrix()

        with self.assertRaises(ValueError):
            a.create_from_omx(temp_file, omx_example, cores=["m1", "m2", "m3", "m4"])

        with self.assertRaises(ValueError):
            a.create_from_omx(temp_file, omx_example, mappings=["wrong index"])
        del a

    def test_get_matrix(self):
        a = AequilibraeMatrix()
        a.load(self.sf_skims)

        with self.assertRaises(AttributeError):
            a.get_matrix("does not exist")

        q = a.get_matrix("distance")
        self.assertEqual(q.shape[0], 24)

        a = AequilibraeMatrix()
        a.load(self.name_test)
        print(np.array_equal(a.get_matrix("seed"), a.matrix["seed"]))

        del a

    def test_save(self):
        a = AequilibraeMatrix()
        a.load(self.sf_skims)

        a.computational_view(["distance"])
        new_mat = np.random.rand(a.zones, a.zones)
        a.matrix_view *= new_mat

        res = a.matrix_view.sum()

        a.save("new_name_for_matrix")
        self.assertEqual(res, a.matrix_view.sum(), "Saved wrong result")

        a.save(["new_name_for_matrix2"])
        self.assertEqual(a.view_names[0], "new_name_for_matrix2", "Did not update computational view")
        self.assertEqual(len(a.view_names), 1, "computational view with the wrong number of matrices")

        a.computational_view(["distance", "new_name_for_matrix"])

        with self.assertRaises(ValueError):
            a.save(["just_one_name"])

        a.save(["one_name", "two_names"])

        with self.assertRaises(ValueError):
            a.save("distance")

        b = AequilibraeMatrix()
        b.load(self.name_test)
        b.computational_view("seed")
        b.save()
        b.computational_view(["mat", "seed", "dist"])
        b.save()


def test_can_reuse_matrix_after_close(tmp_path):
    kwargs = {
        "file_name": tmp_path / "matrix.aem",
        "zones": 1,
        "matrix_names": ["mat"],
        "index_names": ["index"],
    }
    matrix = AequilibraeMatrix()
    matrix.create_empty(**kwargs)
    matrix.close()
    try:
        AequilibraeMatrix().create_empty(**kwargs)
    except OSError:
        pytest.fail("Should not raise OSError")


def test_matrix_reference_doesnt_prevent_resource_cleanup(tmp_path):
    kwargs = {
        "file_name": tmp_path / "matrix.aem",
        "zones": 1,
        "matrix_names": ["mat"],
        "index_names": ["index"],
    }
    matrix = AequilibraeMatrix()
    matrix.create_empty(**kwargs)
    _ = matrix.mat
    del matrix

    try:
        AequilibraeMatrix().create_empty(**kwargs)
    except OSError:
        pytest.fail("Should not raise OSError")



from os import unlink
from os.path import isfile, join
from aequilibrae.project.network.safe_class import SafeClass
from aequilibrae.matrix.aequilibrae_matrix import AequilibraeMatrix


class MatrixRecord(SafeClass):
    def __init__(self, data_set: dict, project):
        super().__init__(data_set, project)
        self._exists = True
        self.fldr = join(project.project_base_path, "matrices")

    def save(self):
        """Saves matrix record to the project database"""
        conn = self.connect_db()
        curr = conn.cursor()

        curr.execute("select count(*) from matrices where name=?", [self.name])
        if curr.fetchone()[0] == 0:
            data = [str(self.name), str(self.file_name), int(self.cores)]
            curr.execute("Insert into matrices (name, file_name, cores) values(?,?,?)", data)

        for key, value in self.__dict__.items():
            if key != "name" and key in self.__original__:
                v_old = self.__original__.get(key, None)
                if value != v_old and value:
                    self.__original__[key] = value
                    curr.execute(f"update matrices set '{key}'=? where name=?", [value, self.name])
        conn.commit()
        conn.close()

    def delete(self):
        """Deletes this matrix record and the underlying data from disk"""
        conn = self.connect_db()
        curr = conn.cursor()
        curr.execute("DELETE FROM matrices where name=?", [self.name])
        conn.commit()
        if isfile(join(self.fldr, self.file_name)):
            try:
                unlink(join(self.fldr, self.file_name))
            except Exception as e:
                self._logger.error(f"Could not remove matrix from disk: {e.args}")

        conn.close()
        self._exists = False

    def update_cores(self):
        """Updates this matrix record with the matrix core count in disk"""
        self.__dict__["cores"] = self.__get_cores()

    def get_data(self) -> AequilibraeMatrix:
        """Returns the actual matrix for further computation

        Returns:
            matrix (:obj:`AequilibraeMatrix`:) Matrix object
        """
        mat = AequilibraeMatrix()
        mat.load(join(self.fldr, self.file_name))
        return mat

    def __setattr__(self, instance, value) -> None:
        if instance == "name":
            value = str(value).lower()
            conn = self.connect_db()
            curr = conn.cursor()
            curr.execute("Select count(*) from matrices where LOWER(name)=?", [value])
            if sum(curr.fetchone()) > 0:
                raise ValueError("Another matrix with this name already exists")
            conn.close()
        elif instance == "file_name":
            conn = self.connect_db()
            curr = conn.cursor()
            curr.execute("Select count(*) from matrices where LOWER(file_name)=?", [str(value).lower()])
            if sum(curr.fetchone()) > 0:
                raise ValueError("There is another matrix record for this file")

        self.__dict__[instance] = value
        if instance in ["file_name", "cores"]:
            self.__dict__["cores"] = self.__get_cores()

    def __get_cores(self) -> int:
        mat = AequilibraeMatrix()
        mat.load(join(self.fldr, self.file_name))
        cores = len(mat.names)
        mat.close()
        del mat
        return cores



from unittest import TestCase
from aequilibrae.paths.AoN import conical, delta_conical
from multiprocessing import cpu_count
import numpy as np


class TestConical(TestCase):
    def test_conial_funtion(self):
        cores = cpu_count()

        alpha = np.zeros(11)
        beta = np.zeros(11)
        fftime = np.zeros(11)
        capacity = np.zeros(11)
        congested_times = np.zeros(11)
        delta = np.zeros(11)

        alpha.fill(9.0)
        beta.fill(1.06)
        fftime.fill(1)
        capacity.fill(1)
        link_flows = np.arange(11).astype(float) * 0.2

        conical(congested_times, link_flows, capacity, fftime, alpha, beta, cores)

        should_be = np.array(
            [
                1,
                1.017609498,
                1.043053698,
                1.092812279,
                1.228923167,
                2,
                4.828923168,
                8.292812282,
                11.8430537,
                15.4176095,
                19.00220725,
            ]
        )

        for i in range(11):
            self.assertAlmostEqual(should_be[i], congested_times[i], 5, "Conical is wrong")

        # Let's check the derivative for sections of the curve
        dx = 0.00000001
        for i in range(1, 11):
            link_flows.fill(1 * 0.2 * i)
            link_flows += np.arange(11) * dx

            conical(congested_times, link_flows, capacity, fftime, alpha, beta, cores)
            delta_conical(delta, link_flows, capacity, fftime, alpha, beta, cores)

            # The derivative needs to be monotonically increasing.
            self.assertGreater(delta[1], delta[0], "Delta is not increasing as it should")

            # We check if the analytical solution matches the numerical differentiation
            dydx = (congested_times[1] - congested_times[0]) / dx
            self.assertAlmostEqual(dydx, delta[1], 6, "Problems with derivative for the conical vdf")
            print(dydx, delta[1])



def randomword(num):
    lorem_ipsum = "Lorem ipsum dolor sit amet consectetur adipiscing elit Maecenas in finibus arcu nec auctor diam Donec porttitor tellus viverra vehicula justo sit amet consectetur arcu Vestibulum id tempus lorem eget pharetra purus Mauris et lorem sit amet erat gravida commodo ac et odio Integer at orci enim Aliquam ut mauris arcu Etiam eget congue nisl Cras orci leo condimentum sit amet porttitor eget pellentesque eget leo Praesent malesuada dictum ullamcorper Nullam malesuada risus eget ligula consectetur sodales Donec cursus volutpat varius"

    lorem_ipsum = lorem_ipsum.split()

    return lorem_ipsum[num]



from random import randint
import pytest

from numpy import array
from shapely.geometry import MultiLineString
from aequilibrae.transit.functions.get_srid import get_srid

from aequilibrae.transit.transit_elements import Route
from tests.aequilibrae.transit.random_word import randomword


class TestRoute:
    @pytest.fixture
    def data_dict(self):
        return {
            "route_id": randomword(randint(0, 40)),
            "route_short_name": randomword(randint(0, 40)),
            "route_long_name": randomword(randint(0, 40)),
            "route_desc": randomword(randint(0, 40)),
            "route_type": randint(0, 13),
            "route_url": randomword(randint(0, 40)),
            "route_color": randomword(randint(0, 40)),
            "route_text_color": randomword(randint(0, 40)),
            "route_sort_order": randint(0, 2000),
            "agency_id": randint(0, 1000),
        }

    def test__populate(self, data_dict):
        data = data_dict

        r = Route(1)
        r.populate(tuple(data.values()), list(data.keys()))
        data["route"] = data.pop("route_id")
        for key, val in r.__dict__.items():
            if key in data:
                assert val == data[key], "Route population with record failed"

        data[randomword(randint(1, 15))] = randomword(randint(1, 20))
        new_r = Route(1)
        with pytest.raises(KeyError):
            new_r.populate(tuple(data.values()), list(data.keys()))

    def test_save_to_database(self, data_dict, transit_conn):
        data = data_dict

        r = Route(1)
        r.srid = get_srid()
        r.populate(tuple(data.values()), list(data.keys()))
        r.shape = MultiLineString([array(((0.0, 0.0), (1.0, 2.0)))])
        r.save_to_database(transit_conn)
        curr = transit_conn.cursor()
        curr.execute(
            "Select agency_id, shortname, longname, description, route_type from routes where route=?",
            [data["route_id"]],
        )
        result = [x for x in curr.fetchone()]
        expected = [
            data["agency_id"],
            data["route_short_name"],
            data["route_long_name"],
            data["route_desc"],
            data["route_type"],
        ]
        assert result == expected, "Saving route to the database failed"
        del curr



import pandas as pd
import numpy as np
cimport cython


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cdef long long _build_compressed_graph(long long[:] link_idx,
                                  long long[:] links_index,
                                  long long[:] link_edge,
                                  long long[:] a_nodes,
                                  long long[:] b_nodes,
                                  signed char[:] directions,
                                  long long link_id_max,
                                  long long[:] simplified_links,
                                  signed char[:] simplified_directions,
                                  long long[:] counts,
                                  long long[:] all_links,
                                  long long[:] compressed_dir,
                                  long long[:] compressed_a_node,
                                  long long[:] compressed_b_node) nogil:
    cdef:
        long long slink = 0
        long long pre_link, n, first_node, lnk, lidx, a_node, b_node
        long ab_dir = 1  # These could definitely be smaller since we only ever use them in conditionals
        long ba_dir = 1
        long drc
        Py_ssize_t i, k

    for i in range(link_edge.shape[0]):
        pre_link = link_edge[i]

        if simplified_links[pre_link] >= 0:
            continue
        ab_dir = 1
        ba_dir = 1
        lidx = link_idx[pre_link]
        a_node = a_nodes[lidx]
        b_node = b_nodes[lidx]
        drc = directions[lidx]
        n = a_node if counts[a_node] == 2 else b_node
        first_node = b_node if counts[a_node] == 2 else a_node

        ab_dir = 0 if (first_node == a_node and drc < 0) or (first_node == b_node and drc > 0) else ab_dir
        ba_dir = 0 if (first_node == a_node and drc > 0) or (first_node == b_node and drc < 0) else ba_dir

        while counts[n] == 2:
            # assert (simplified_links[pre_link] >= 0), "How the heck did this happen?"
            simplified_links[pre_link] = slink
            simplified_directions[pre_link] = -1 if a_node == n else 1

            # Gets the link from the list that is not the link we are coming from
            for k in range(links_index[n], links_index[n + 1]):
                if pre_link != all_links[k]:
                    pre_link = all_links[k]
                    break

            lidx = link_idx[pre_link]
            a_node = a_nodes[lidx]
            b_node = b_nodes[lidx]
            drc = directions[lidx]
            ab_dir = 0 if (n == a_node and drc < 0) or (n == b_node and drc > 0) else ab_dir
            ba_dir = 0 if (n == a_node and drc > 0) or (n == b_node and drc < 0) else ba_dir
            n = (
                a_nodes[lidx]
                if n == b_nodes[lidx]
                else b_nodes[lidx]
            )

        simplified_links[pre_link] = slink
        simplified_directions[pre_link] = -1 if a_node == n else 1
        last_node = b_node if counts[a_node] == 2 else a_node
        # major_nodes[slink] = [first_node, last_node]

        # Available directions are NOT indexed like the other arrays
        compressed_a_node[slink] = first_node
        compressed_b_node[slink] = last_node
        if ab_dir > 0:
            if ba_dir > 0:
                compressed_dir[slink] = 0
            else:
                compressed_dir[slink] = 1
        elif ba_dir > 0:
            compressed_dir[slink] = -1
        else:
            compressed_dir[slink] = -999
            slink += 1

    return slink


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cdef void _back_fill(long long[:] links_index, long long max_node):
    cdef Py_ssize_t i

    for i in range(max_node + 1, 0, -1):
        links_index[i - 1] = links_index[i] if links_index[i - 1] == -1 else links_index[i - 1]


def build_compressed_graph(graph):
    # Build link index
    link_id_max = graph.network.link_id.max()

    link_idx = np.empty(link_id_max + 1, dtype=np.int64)
    link_idx[graph.network.link_id] = np.arange(graph.network.shape[0])

    nodes = np.hstack([graph.network.a_node.values, graph.network.b_node.values])
    links = np.hstack([graph.network.link_id.values, graph.network.link_id.values])
    counts = np.bincount(nodes) # index (node) i has frequency counts[i]. This is just the number of
    # edges that connection to a given node

    idx = np.argsort(nodes)
    all_nodes = nodes[idx]
    all_links = links[idx]
    all_nodes_max = all_nodes.max()

    links_index = np.full(all_nodes_max + 2, -1, dtype=np.int64)
    nlist = np.arange(all_nodes_max + 2)

    y, x, _ = np.intersect1d(all_nodes, nlist, assume_unique=False, return_indices=True)
    links_index[y] = x[:]
    links_index[-1] = all_links.shape[0]

    _back_fill(links_index[:], all_nodes_max)

    # We keep all centroids for sure
    counts[graph.centroids] = 999

    truth = (counts == 2).astype(np.uint8)
    link_edge = truth[graph.network.a_node.values] + truth[graph.network.b_node.values]
    link_edge = graph.network.link_id.values[link_edge == 1].astype(np.int64)

    simplified_links = np.full(link_id_max + 1, -1, dtype=np.int64)
    simplified_directions = np.zeros(link_id_max + 1, dtype=np.int8)

    compressed_dir = np.zeros(link_id_max + 1, dtype=np.int64)
    compressed_a_node = np.zeros(link_id_max + 1, dtype=np.int64)
    compressed_b_node = np.zeros(link_id_max + 1, dtype=np.int64)

    slink = _build_compressed_graph(
        link_idx[:],
        links_index[:],
        link_edge[:],
        graph.network.a_node.values[:],
        graph.network.b_node.values[:],
        graph.network.direction.values[:],
        link_id_max,
        simplified_links[:],
        simplified_directions[:],
        counts[:],
        all_links[:],
        compressed_dir[:],
        compressed_a_node[:],
        compressed_b_node[:]
    )

    links_to_remove = np.argwhere(simplified_links >= 0)
    df = pd.DataFrame(graph.network, copy=True)
    if links_to_remove.shape[0]:
        df = df[~df.link_id.isin(links_to_remove[:, 0])]
        df = df[df.a_node != df.b_node]

    comp_lnk = pd.DataFrame(
        {
            "a_node": compressed_a_node[:slink],
            "b_node": compressed_b_node[:slink],
            "direction": compressed_dir[:slink],
            "link_id": np.arange(slink),
        }
    )
    max_link_id = link_id_max * 10
    comp_lnk.loc[:, "link_id"] += max_link_id

    df = pd.concat([df, comp_lnk])
    df = df[["id", "link_id", "a_node", "b_node", "direction"]]
    properties = graph._Graph__build_directed_graph(df, graph.centroids)  # FIXME: Don't circumvent name mangling
    graph.compact_all_nodes = properties[0]
    graph.compact_num_nodes = properties[1]
    graph.compact_nodes_to_indices = properties[2]
    graph.compact_fs = properties[3]
    graph.compact_graph = properties[4]

    crosswalk = pd.DataFrame(
        {
            "link_id": np.arange(simplified_directions.shape[0]),
            "link_direction": simplified_directions,
            "compressed_link": simplified_links,
            "compressed_direction": np.ones(simplified_directions.shape[0], dtype=np.int8),
        }
    )

    crosswalk = crosswalk[crosswalk.compressed_link >= 0]
    crosswalk.loc[:, "compressed_link"] += max_link_id

    cw2 = pd.DataFrame(crosswalk, copy=True)
    cw2.loc[:, "link_direction"] *= -1
    cw2.loc[:, "compressed_direction"] = -1

    crosswalk = pd.concat([crosswalk, cw2])
    crosswalk = crosswalk.assign(key=crosswalk.compressed_link * crosswalk.compressed_direction)
    crosswalk.drop(["compressed_link", "compressed_direction"], axis=1, inplace=True)

    final_ids = pd.DataFrame(graph.compact_graph[["id", "link_id", "direction"]], copy=True)
    final_ids = final_ids.assign(key=final_ids.link_id * final_ids.direction)
    final_ids.drop(["link_id", "direction"], axis=1, inplace=True)

    agg_crosswalk = crosswalk.merge(final_ids, on="key")
    agg_crosswalk.loc[:, "key"] = agg_crosswalk.link_id * agg_crosswalk.link_direction
    agg_crosswalk.drop(["link_id", "link_direction"], axis=1, inplace=True)

    direct_crosswalk = final_ids[final_ids.key.abs() < max_link_id]

    crosswalk = pd.concat([agg_crosswalk, direct_crosswalk])[["key", "id"]]
    crosswalk.columns = ["__graph_correlation_key__", "__compressed_id__"]

    graph.graph = graph.graph.assign(__graph_correlation_key__=graph.graph.link_id * graph.graph.direction)
    graph.graph = graph.graph.merge(crosswalk, on="__graph_correlation_key__", how="left")
    graph.graph.drop(["__graph_correlation_key__"], axis=1, inplace=True)

    # If will refer all the links that have no correlation to an element beyond the last link
    # This element will always be zero during assignment
    graph.graph.loc[graph.graph.__compressed_id__.isna(), "__compressed_id__"] = graph.compact_graph.id.max() + 1



from aequilibrae.project import Project
from aequilibrae.project.database_connection import database_connection
import pytest


class TestProject:
    def test_opening_wrong_folder(self, tmp_path):
        not_a_project = str(tmp_path)
        with pytest.raises(FileNotFoundError):
            proj = Project()
            proj.open(not_a_project)

    @pytest.mark.parametrize(
        "table, exp_column",
        [
            ("links", "distance"),
            ("nodes", "is_centroid"),
        ],
    )
    def test_table_creation(self, table: str, exp_column: str, project: Project):
        curr = project.conn.cursor()
        curr.execute(f"PRAGMA table_info({table});")
        fields = curr.fetchall()
        fields = {x[1] for x in fields}

        assert exp_column in fields, f"Table {table.upper()} was not created correctly"

    def test_close(self, project):
        database_connection(db_type="network")

        project.close()
        with pytest.raises(FileNotFoundError):
            database_connection(db_type="network")



from unittest import TestCase
from sqlite3 import IntegrityError
import os
from shutil import copytree, rmtree
from random import randint, random
import uuid
from tempfile import gettempdir
from shapely.geometry import Point
import shapely.wkb
from aequilibrae.project import Project

from ...data import siouxfalls_project


class TestNode(TestCase):
    def setUp(self) -> None:
        os.environ["PATH"] = os.path.join(gettempdir(), "temp_data") + ";" + os.environ["PATH"]

        self.proj_dir = os.path.join(gettempdir(), uuid.uuid4().hex)
        copytree(siouxfalls_project, self.proj_dir)

        self.project = Project()
        self.project.open(self.proj_dir)
        self.network = self.project.network
        self.curr = self.project.conn.cursor()

    def tearDown(self) -> None:
        self.curr.close()
        self.project.close()
        try:
            rmtree(self.proj_dir)
        except Exception as e:
            print(f"Failed to remove at {e.args}")

    def test_save_and_assignment(self):
        nodes = self.network.nodes
        nd = randint(1, 24)
        node = nodes.get(nd)

        with self.assertRaises(AttributeError):
            node.modes = "abc"

        with self.assertRaises(AttributeError):
            node.link_types = "default"

        with self.assertRaises(AttributeError):
            node.node_id = 2

        with self.assertRaises(ValueError):
            node.is_centroid = 2

        node.is_centroid = 0
        self.assertEqual(0, node.is_centroid, "Assignment of is_centroid did not work")

        x = node.geometry.x + random()
        y = node.geometry.y + random()

        node.geometry = Point([x, y])

        node.save()

        self.curr.execute("Select is_centroid, asBinary(geometry) from nodes where node_id=?;", [nd])
        flag, wkb = self.curr.fetchone()
        self.assertEqual(flag, 0, "Saving of is_centroid failed")

        geo = shapely.wkb.loads(wkb)
        self.assertEqual(geo.x, x, "Geometry X saved wrong")
        self.assertEqual(geo.y, y, "Geometry Y saved wrong")

        self.curr.execute("Select asBinary(geometry) from links where a_node=?;", [nd])
        wkb = self.curr.fetchone()[0]

        geo2 = shapely.wkb.loads(wkb)
        self.assertEqual(geo2.xy[0][0], x, "Saving node geometry broke underlying network")
        self.assertEqual(geo2.xy[1][0], y, "Saving node geometry broke underlying network")

    def test_data_fields(self):
        nodes = self.network.nodes

        node1 = nodes.get(randint(1, 24))
        node2 = nodes.get(randint(1, 24))

        self.assertEqual(node1.data_fields(), node2.data_fields(), "Different nodes have different data fields")

        fields = sorted(node1.data_fields())
        self.curr.execute("pragma table_info(nodes)")
        dt = self.curr.fetchall()

        actual_fields = sorted([x[1] for x in dt if x[1] != "ogc_fid"])

        self.assertEqual(fields, actual_fields, "Node has unexpected set of fields")

    def test_renumber(self):
        nodes = self.network.nodes

        node = nodes.get(randint(2, 24))

        x = node.geometry.x
        y = node.geometry.y

        with self.assertRaises(IntegrityError):
            node.renumber(1)

        num = randint(25, 2000)
        node.renumber(num)

        self.curr.execute("Select asBinary(geometry) from nodes where node_id=?;", [num])
        wkb = self.curr.fetchone()[0]

        geo = shapely.wkb.loads(wkb)
        self.assertEqual(geo.x, x, "Renumbering failed")
        self.assertEqual(geo.y, y, "Renumbering failed")



import pytest
from datetime import datetime, timedelta
import time
from random import randint
from aequilibrae.transit.date_tools import to_seconds, to_time_string, one_day_before, create_days_between
from aequilibrae.transit.date_tools import day_of_week, format_date


class TestDateTools:
    def test_to_seconds(self):
        t = to_seconds("18:45:36")
        assert t == (18 * 3600 + 45 * 60 + 36), "to_seconds failed to return right values"

        t = to_seconds("1:5:6")
        assert t == (1 * 3600 + 5 * 60 + 6), "to_seconds failed to return right values"

        with pytest.raises(ValueError):
            t = to_seconds("")

        t = to_seconds("1 day, 1:5:6")
        assert t == (25 * 3600 + 5 * 60 + 6), "to_seconds failed to return right values"

        t = to_seconds("2 day, 1:5:6")
        assert t == (49 * 3600 + 5 * 60 + 6), "to_seconds failed to return right values"

        t = to_seconds("-1 day, 1:5:6")
        assert t == (-23 * 3600 + 5 * 60 + 6), "to_seconds failed to return right values"

    def test_to_time_string(self):
        with pytest.raises(TypeError):
            _ = to_time_string("18:45:36")

        t = to_time_string(None)
        assert t is None, "to_time_string failed to return None"

        t = to_time_string(86399)
        assert t == "23:59:59", "to_time_string failed to return right values"

        t = to_time_string("86398")
        assert t == "23:59:58", "to_time_string failed to return right values"

        t = to_time_string(86401)
        assert t == "1 day, 0:00:01", "to_time_string failed to return right values"

        t = to_seconds("2 day, 1:5:6")
        assert t == (49 * 3600 + 5 * 60 + 6), "to_seconds failed to return right values"

    def test_one_day_before(self):
        t = one_day_before(datetime.fromisoformat("2020-01-01"))
        assert t == "2019-12-31"

        t = one_day_before(datetime.fromisoformat("2020-03-01"))
        assert t == "2020-02-29"

        t = one_day_before(datetime.fromisoformat("2020-02-29"))
        assert t == "2020-02-28"

        with pytest.raises(ValueError):
            _ = one_day_before(datetime.fromisoformat("2020-13-01"))

        with pytest.raises(TypeError):
            _ = one_day_before(datetime.fromisoformat(123456))

    def test_create_days_between(self):
        tdy = datetime.today()
        days = randint(1, 100)
        past = tdy - timedelta(days=days)

        interval = create_days_between(past, tdy)
        assert len(interval) == days + 1, "create_days_between returned wrong value"

        interval = create_days_between(tdy, past)
        assert len(interval) == 0, "create_days_between returned wrong value"

        interval = create_days_between(tdy, tdy)
        assert len(interval) == 1, "create_days_between returned wrong value"

    def test_day_of_week(self):
        tdy = datetime.today()
        days = randint(1, 100)
        past = tdy - timedelta(days=days)

        old = f"{past.year}-{past.month:02d}-{past.day:02d}"
        today = f"{tdy.year}-{tdy.month:02d}-{tdy.day:02d}"

        for date in [today, old]:
            should = time.strptime(date, "%Y-%m-%d").tm_wday
            actually = day_of_week(date)
            assert should == actually, "day_of_week returned the wrong value"

        with pytest.raises(TypeError):
            day_of_week(100)

    def test_format_date(self):
        for i in range(100):
            year = randint(1000, 2020)
            month = randint(1, 12)
            day = randint(1, 28)
            date = f"{year}{month:02d}{day:02d}"
            formatted = format_date(date)
            assert formatted == f"{year}-{month:02d}-{day:02d}"



"""
Original Algorithm for Shortest path (Dijkstra with a 4-ary heap) was written by François Pacull <francois.pacull@architecture-performance.fr> under license: MIT, (C) 2022
"""

"""
TODO:
LIST OF ALL THE THINGS WE NEED TO DO TO NOT HAVE TO HAVE nodes 1..n as CENTROIDS. ARBITRARY NUMBERING
- Checks of weather the centroid we are computing path from is a centroid and/or exists in the graph
- Re-write function **network_loading** on the part of loading flows to centroids
"""
cimport cython
from libc.math cimport isnan, INFINITY
from libc.string cimport memset

include 'pq_4ary_heap.pyx'

@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False) # turn off bounds-checking for entire function
cpdef void network_loading(long classes,
                           double[:, :] demand,
                           long long [:] pred,
                           long long [:] conn,
                           double[:, :] link_loads,
                           long long [:] no_path,
                           long long [:] reached_first,
                           double [:, :] node_load,
                           long found) nogil:

    cdef long long i, j, predecessor, connector, node
    cdef long long zones = demand.shape[0]
    cdef long long N = node_load.shape[0]
# Traditional loading, without cascading
#    for i in range(zones):
#        node = i
#        predecessor = pred[node]
#        connector = conn[node]
#        while predecessor >= 0:
#            for j in range(classes):
#                link_loads[connector, j] += demand[i, j]
#
#            predecessor = pred[predecessor]
#            connector = conn[predecessor]

    # Clean the node load array
    for i in range(N):
        node_load[i] = 0

    # Loads the demand to the centroids
    for j in range(classes):
        for i in range(zones):
            if not isnan(demand[i, j]):
                node_load[i, j] = demand[i, j]

    #Recursively cascades to the origin
    for i in range(found, 0, -1):
        node = reached_first[i]

        # captures how we got to that node
        predecessor = pred[node]
        connector = conn[node]

        # loads the flow to the links for each class
        for j in range(classes):
            link_loads[connector, j] += node_load[node, j]
            # Cascades the load from the node to their predecessor
            node_load[predecessor, j] += node_load[node, j]

@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cdef void _copy_skims(double[:,:] skim_matrix,  #Skim matrix_procedures computed from one origin to all nodes
                      double[:,:] final_skim_matrix) nogil:  #Skim matrix_procedures computed for one origin to all other centroids only

    cdef long i, j
    cdef long N = final_skim_matrix.shape[0]
    cdef long skims = final_skim_matrix.shape[1]

    for i in range(N):
        for j in range(skims):
            final_skim_matrix[i,j]=skim_matrix[i,j]


cdef return_an_int_view(input):
    cdef int [:] critical_links_view = input
    return critical_links_view

@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cdef void sl_network_loading(
    long long [:, :] selected_links,
    double [:, :] demand,
    long long [:] pred,
    long long [:] conn,
    double [:, :] link_loads,
    double [:, :, :] sl_od_matrix,
    double [:, :, :] sl_link_loading,
    unsigned char [:] has_flow_mask,
    long classes) nogil:
# VARIABLES:
#   selected_links: 2d memoryview. Each row corresponds to a set of selected links specified by the user
#   demand:         The input demand matrix for a given origin. The first index corresponds to destination,
#                   second is the class
#   pred:           The list of predecessor nodes, i.e. given a node, referencing that node's index in pred
#                   yields the previous node in the minimum spanning tree
#   conn:           The list of links which connect predecessor nodes. referencing it by the predecessor yields
#                   the link it used to connect the two nodes
# link_loads:       Stores the loading on each link. Equivalent to link_loads in network_loading
# temp_sl_od_matrix:     Stores the OD matrix for each set of selected links sliced for the given origin
# The indices are:  set of links, destination, class
# temp_sl_link_loading:  Stores the loading on the Selected links, and the paths which use the selected links
#                   The indices are: set of links, link_id, class)
# has_flow_mask:    An array which acts as a flag for which links were used in retracing a given OD path
# classes:          the number of subclasses of vehicles for the given TrafficClass
# 
# Executes regular loading, while keeping track of SL links
    cdef:
        int i, j, k, l, dests = demand.shape[0], xshape = has_flow_mask.shape[0]
        long long predecessor, connection, lid, link
        bint found
    for j in range(dests):
        memset(&has_flow_mask[0], 0, xshape * sizeof(unsigned char))
        connection = conn[j]
        predecessor = pred[j]

        # Walk the path and mark all used links in the has_flow_mask
        while predecessor >= 0:
            for k in range(classes):
                link_loads[connection, k] += demand[j, k]
            has_flow_mask[connection] = 1
            connection = conn[predecessor]
            predecessor = pred[predecessor]
        # Now iterate through each SL set and see if any of their links were marked
        for i in range(selected_links.shape[0]):
            # We check to see if the given OD path marked any of our selected links
            found = 0
            l = 0
            while l < selected_links.shape[1] and found == 0:
                # Checks to see if the current set of selected links has finished
                # NOTE: -1 is a default value for the selected_links array. It cannot be a link id, if -1 turns up
                # There is either a serious bug, or the program has reached the end of a set of links in SL.
                # This lets us early exit from the loop without needing to iterate through the rest of the array
                # Which just has placeholder values
                if selected_links[i][l] == -1:
                    break
                if has_flow_mask[selected_links[i][l]] != 0:
                    found = 1
                l += 1
            if found == 0:
                continue
            for k in range(classes):
                sl_od_matrix[i, j, k] = demand[j, k]
            connection = conn[j]
            predecessor = pred[j]
            while predecessor >= 0:
                for k in range(classes):
                    sl_link_loading[i, connection, k] += demand[j, k]
                connection = conn[predecessor]
                predecessor = pred[predecessor]

@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cpdef void put_path_file_on_disk(unsigned int orig,
                                 unsigned int [:] pred,
                                 long long [:] predecessors,
                                 unsigned int [:] conn,
                                 long long [:] connectors,
                                 long long [:] all_nodes,
                                 unsigned int [:] origins_to_write,
                                 unsigned int [:] nodes_to_write) nogil:
    cdef long long i
    cdef long long k = pred.shape[0]

    for i in range(k):
        origins_to_write[i] = orig
        nodes_to_write[i] = all_nodes[i]
        pred[i] = all_nodes[predecessors[i]]
        conn[i] = connectors[i]


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cdef void blocking_centroid_flows(int action,
                                  long long orig,
                                  long long centroids,
                                  long long [:] fs,
                                  long long [:] temp_b_nodes,
                                  long long [:] real_b_nodes) nogil:
    cdef long long i

    if action == 1: # We are unblocking
        for i in range(fs[centroids]):
            temp_b_nodes[i] = real_b_nodes[i]
    else: # We are blocking:
        for i in range(fs[centroids]):
            temp_b_nodes[i] = orig

        for i in range(fs[orig], fs[orig + 1]):
            temp_b_nodes[i] = real_b_nodes[i]


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False) # turn of bounds-checking for entire function
cdef void skim_single_path(long origin,
                           long nodes,
                           long skims,
                           double[:, :] node_skims,
                           long long [:] pred,
                           long long [:] conn,
                           double[:, :] graph_costs,
                           long long [:] reached_first,
                           long found) nogil:
    cdef long long i, node, predecessor, connector, j

    # sets all skims to infinity
    for i in range(nodes):
        for j in range(skims):
            node_skims[i, j] = INFINITY

    # Zeroes the intrazonal cost
    for j in range(skims):
            node_skims[origin, j] = 0

    # Cascade skimming
    for i in range(1, found + 1):
        node = reached_first[i]

        # captures how we got to that node
        predecessor = pred[node]
        connector = conn[node]

        for j in range(skims):
            node_skims[node, j] = node_skims[predecessor, j] + graph_costs[connector, j]


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False) # turn of bounds-checking for entire function
cpdef void skim_multiple_fields(long origin,
                                long nodes,
                                long zones,
                                long skims,
                                double[:, :] node_skims,
                                long long [:] pred,
                                long long [:] conn,
                                double[:, :] graph_costs,
                                long long [:] reached_first,
                                long found,
                                double [:,:] final_skims) nogil:
    cdef long long i, node, predecessor, connector, j

    # sets all skims to infinity
    for i in range(nodes):
        for j in range(skims):
            node_skims[i, j] = INFINITY

    # Zeroes the intrazonal cost
    for j in range(skims):
            node_skims[origin, j] = 0

    # Cascade skimming
    for i in range(1, found + 1):
        node = reached_first[i]

        # captures how we got to that node
        predecessor = pred[node]
        connector = conn[node]

        for j in range(skims):
            node_skims[node, j] = node_skims[predecessor, j] + graph_costs[connector, j]

    for i in range(zones):
        for j in range(skims):
            final_skims[i, j] = node_skims[i, j]

# ###########################################################################################################################
#############################################################################################################################
# Original Dijkstra implementation by François Pacull, taken from https://github.com/Edsger-dev/priority_queues
# Old Numpy Buffers were replaces with latest memory views interface to allow for the release of the GIL
# Path tracking arrays and skim arrays were also added to it
#############################################################################################################################
# ###########################################################################################################################

@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False) # turn of bounds-checking for entire function
cpdef int path_finding(long origin,
                       double[:] graph_costs,
                       long long [:] csr_indices,
                       long long [:] graph_fs,
                       long long [:] pred,
                       long long [:] ids,
                       long long [:] connectors,
                       long long [:] reached_first) nogil:

    cdef unsigned int N = graph_costs.shape[0]
    cdef unsigned int M = pred.shape[0]


    cdef:
        size_t tail_vert_idx, head_vert_idx, idx  # indices
        DTYPE_t tail_vert_val, head_vert_val  # vertex travel times
        PriorityQueue pqueue  # binary heap
        ElementState vert_state  # vertex state
        size_t origin_vert = <size_t>origin
        ITYPE_t found = 0

    for i in range(M):
        pred[i] = -1
        connectors[i] = -1
        reached_first[i] = -1

    # initialization of the heap elements
    # all nodes have INFINITY key and NOT_IN_HEAP state
    init_heap(&pqueue, <size_t>N)

    # the key is set to zero for the origin vertex,
    # which is inserted into the heap
    insert(&pqueue, origin_vert, 0.0)

    # main loop
    while pqueue.size > 0:
        tail_vert_idx = extract_min(&pqueue)
        tail_vert_val = pqueue.Elements[tail_vert_idx].key
        reached_first[found] = tail_vert_idx
        found += 1

        # loop on outgoing edges
        for idx in range(<size_t>graph_fs[tail_vert_idx], <size_t>graph_fs[tail_vert_idx + 1]):
            head_vert_idx = <size_t>csr_indices[idx]
            vert_state = pqueue.Elements[head_vert_idx].state
            if vert_state != SCANNED:
                head_vert_val = tail_vert_val + graph_costs[idx]
                if vert_state == NOT_IN_HEAP:
                    insert(&pqueue, head_vert_idx, head_vert_val)
                    pred[head_vert_idx] = tail_vert_idx
                    connectors[head_vert_idx] = ids[idx]
                elif pqueue.Elements[head_vert_idx].key > head_vert_val:
                    decrease_key(&pqueue, head_vert_idx, head_vert_val)
                    pred[head_vert_idx] = tail_vert_idx
                    connectors[head_vert_idx] = ids[idx]

    free_heap(&pqueue)
    return found - 1



from unittest import TestCase
from aequilibrae.project.network.osm_utils.place_getter import placegetter
from random import random
import os


class Test(TestCase):
    def test_placegetter(self):
        thresh = 0.05
        if os.environ.get("GITHUB_WORKFLOW", "ERROR") == "Code coverage":
            thresh = 1.01

        if random() < thresh:
            place, report = placegetter("Vatican City")
            place = [round(x, 1) for x in place]
            if place != [12.4, 41.9, 12.5, 41.9]:
                self.fail("Returned the wrong boundingbox for Vatican City")

            place, report = placegetter("Just a random place with no bear in reality")
            if place is not None:
                self.fail("Returned a bounding box for a place that does not exist")
        else:
            print("Skipped check to not load OSM servers")



from logging import FileHandler
import os
import uuid
from shutil import copytree
from tempfile import gettempdir
from unittest import TestCase

import pytest

from ...data import siouxfalls_project
from aequilibrae.project import Project
from aequilibrae import log
from aequilibrae.log import _setup_logger


class TestLog(TestCase):
    def setUp(self) -> None:
        self.proj_dir = os.path.join(gettempdir(), uuid.uuid4().hex)
        copytree(siouxfalls_project, self.proj_dir)

        self.project = Project()
        self.project.open(self.proj_dir)

    def tearDown(self) -> None:
        self.project.close()

    def test_contents(self):
        log = self.project.log()
        cont = log.contents()
        self.assertEqual(len(cont), 4, "Returned the wrong amount of data from the log")

    def test_clear(self):
        log = self.project.log()
        log.clear()

        with open(os.path.join(self.proj_dir, "aequilibrae.log"), "r") as file:
            q = file.readlines()
        self.assertEqual(len(q), 0, "Failed to clear the log file")


class TestStartsLogging:
    @pytest.fixture
    def project(self, create_project):
        return create_project()

    @staticmethod
    def get_handlers(logger, name="aequilibrae"):
        return [h for h in logger.handlers if h.name == name]

    @classmethod
    def get_logger_file(cls, logger, name="aequilibrae") -> str:
        handlers = cls.get_handlers(logger, name)
        if not handlers:
            raise ValueError(f"Logger has no handlers named {name}")
        handler = handlers[0]
        if not isinstance(handler, FileHandler):
            raise TypeError(f"Handler must be FileHandler, not {type(handler).__name__}")

        return handlers[0].baseFilename

    def test_logger_has_single_handler_named_aquilibrae(self):
        _setup_logger()
        logger = _setup_logger()
        assert len(self.get_handlers(logger)) == 1

    def test_default_logger_handler(self):
        logger = _setup_logger()
        assert self.get_logger_file(logger).endswith("aequilibrae.log")

    def test_project_logger(self, create_project):
        project = create_project()
        assert self.get_logger_file(project.logger).startswith(project.project_base_path)

    def test_activate_project_leaves_global_logger_intact(self, create_project):
        project = create_project()
        assert self.get_logger_file(log.global_logger) != self.get_logger_file(project.logger)

    def test_multiple_projects_have_separate_logger(self, create_project):
        a = create_project()
        b = create_project()
        assert self.get_logger_file(a.logger) != self.get_logger_file(b.logger)



import os
import sys
import uuid
import zipfile
from os.path import join
from shutil import copytree
from tempfile import gettempdir
from unittest import TestCase
from uuid import uuid4

import numpy as np

from aequilibrae import Project
from aequilibrae.paths import path_computation, Graph
from aequilibrae.paths.results import PathResults
from aequilibrae.utils.create_example import create_example
from ...data import triangle_graph_blocking, st_varent_network

# Adds the folder with the data to the path and collects the paths to the files
lib_path = os.path.abspath(os.path.join("..", "../tests"))
sys.path.append(lib_path)

origin = 5
dest = 13


class TestPathResults(TestCase):
    def setUp(self) -> None:
        self.project = create_example(join(gettempdir(), "test_set_pce_" + uuid4().hex))
        self.project.network.build_graphs()
        self.g = self.project.network.graphs["c"]  # type: Graph
        self.g.set_graph("free_flow_time")
        self.g.set_blocked_centroid_flows(False)

        self.matrix = self.project.matrices.get_matrix("demand_omx")
        self.matrix.computational_view()

        self.r = PathResults()
        self.r.prepare(self.g)

    def tearDown(self) -> None:
        self.project.close()
        self.matrix.close()
        del self.r

    def test_reset(self):
        self.r.compute_path(dest, origin)
        self.r.reset()

        self.assertEqual(self.r.path, None, "Fail to reset the Path computation object")
        self.assertEqual(self.r.path_nodes, None, "Fail to reset the Path computation object")
        self.assertEqual(self.r.path_link_directions, None, "Fail to reset the Path computation object")
        self.assertEqual(self.r.milepost, None, "Fail to reset the Path computation object")
        self.assertEqual(self.r.predecessors.max(), -1, "Fail to reset the Path computation object")
        self.assertEqual(self.r.predecessors.min(), -1, "Fail to reset the Path computation object")
        self.assertEqual(self.r.connectors.max(), -1, "Fail to reset the Path computation object")
        self.assertEqual(self.r.connectors.min(), -1, "Fail to reset the Path computation object")
        if self.r.skims is not None:
            self.assertEqual(self.r.skims.max(), np.inf, "Fail to reset the Path computation object")
            self.assertEqual(self.r.skims.min(), np.inf, "Fail to reset the Path computation object")

        new_r = PathResults()
        with self.assertRaises(ValueError):
            new_r.reset()

    def test_compute_paths(self):
        path_computation(5, 2, self.g, self.r)

        self.assertEqual(list(self.r.path), [12, 14], "Path computation failed. Wrong sequence of links")
        self.assertEqual(list(self.r.path_link_directions), [1, 1], "Path computation failed. Wrong link directions")
        self.assertEqual(list(self.r.path_nodes), [5, 6, 2], "Path computation failed. Wrong sequence of path nodes")
        self.assertEqual(list(self.r.milepost), [0, 4, 9], "Path computation failed. Wrong milepost results")

    def test_compute_with_skimming(self):
        r = PathResults()
        self.g.set_skimming("free_flow_time")
        r.prepare(self.g)
        r.compute_path(origin, dest)
        self.assertEqual(r.milepost[-1], r.skims[dest], "Skims computed wrong when computing path")

    def test_update_trace(self):
        self.r.compute_path(origin, 2)

        self.r.update_trace(10)
        self.assertEqual(list(self.r.path), [13, 25], "Path update failed. Wrong sequence of links")
        self.assertEqual(list(self.r.path_link_directions), [1, 1], "Path update failed. Wrong link directions")
        self.assertEqual(list(self.r.path_nodes), [5, 9, 10], "Path update failed. Wrong sequence of path nodes")
        self.assertEqual(list(self.r.milepost), [0, 5, 8], "Path update failed. Wrong milepost results")


class TestBlockingTrianglePathResults(TestCase):
    def setUp(self) -> None:
        os.environ["PATH"] = os.path.join(gettempdir(), "temp_data") + ";" + os.environ["PATH"]
        self.proj_dir = os.path.join(gettempdir(), uuid.uuid4().hex)
        copytree(triangle_graph_blocking, self.proj_dir)
        self.project = Project()
        self.project.open(self.proj_dir)
        self.project.network.build_graphs(modes=["c"])
        self.g = self.project.network.graphs["c"]  # type: Graph
        self.g.set_graph("free_flow_time")
        self.g.set_blocked_centroid_flows(True)

        self.r = PathResults()
        self.r.prepare(self.g)

    def tearDown(self) -> None:
        self.project.close()
        del self.r

    def test_compute_paths(self):
        self.r.compute_path(1, 2)
        self.assertEqual(list(self.r.path_nodes), [1, 3, 2])
        self.assertEqual(list(self.r.path), [1, 2])

        self.r.compute_path(2, 1)
        self.assertEqual(list(self.r.path_nodes), [2, 1])
        self.assertEqual(list(self.r.path), [3])

        self.r.compute_path(3, 1)
        self.assertEqual(list(self.r.path_nodes), [3, 2, 1])
        self.assertEqual(list(self.r.path), [2, 3])

        self.r.compute_path(3, 2)
        self.assertEqual(list(self.r.path_nodes), [3, 2])
        self.assertEqual(list(self.r.path), [2])

        self.r.compute_path(1, 3)
        self.assertEqual(list(self.r.path_nodes), [1, 3])
        self.assertEqual(list(self.r.path), [1])

        self.r.compute_path(2, 3)
        self.assertEqual(list(self.r.path_nodes), [2, 1, 3])
        self.assertEqual(list(self.r.path), [3, 1])

    def test_compute_blocking_paths(self):
        self.r.compute_path(4, 5)
        self.assertEqual(list(self.r.path_nodes), [4, 1, 3, 2, 5])
        self.assertEqual(list(self.r.path), [4, 1, 2, 5])

        self.r.compute_path(5, 4)
        self.assertEqual(list(self.r.path_nodes), [5, 2, 1, 4])
        self.assertEqual(list(self.r.path), [5, 3, 4])

        self.r.compute_path(6, 4)
        self.assertEqual(list(self.r.path_nodes), [6, 3, 2, 1, 4])
        self.assertEqual(list(self.r.path), [6, 2, 3, 4])

        self.r.compute_path(6, 5)
        self.assertEqual(list(self.r.path_nodes), [6, 3, 2, 5])
        self.assertEqual(list(self.r.path), [6, 2, 5])

        self.r.compute_path(4, 6)
        self.assertEqual(list(self.r.path_nodes), [4, 1, 3, 6])
        self.assertEqual(list(self.r.path), [4, 1, 6])

        self.r.compute_path(5, 6)
        self.assertEqual(list(self.r.path_nodes), [5, 2, 1, 3, 6])
        self.assertEqual(list(self.r.path), [5, 3, 1, 6])

    def test_update_trace(self):
        self.r.compute_path(1, 2)
        self.assertEqual(list(self.r.path_nodes), [1, 3, 2])
        self.assertEqual(list(self.r.path), [1, 2])

        self.r.update_trace(3)
        self.assertEqual(list(self.r.path_nodes), [1, 3])
        self.assertEqual(list(self.r.path), [1])

    def test_update_blocking_trace(self):
        self.r.compute_path(4, 5)
        self.assertEqual(list(self.r.path_nodes), [4, 1, 3, 2, 5])
        self.assertEqual(list(self.r.path), [4, 1, 2, 5])

        self.r.update_trace(6)
        self.assertEqual(list(self.r.path_nodes), [4, 1, 3, 6])
        self.assertEqual(list(self.r.path), [4, 1, 6])


class TestCentroidsLast(TestCase):
    def test_compute_paths_centroid_last_node_id(self):
        # Issue 307 consisted of not being able to compute paths between two
        # centroids if we were skimming and one of them had the highest node_id
        # in the entire network
        zipfile.ZipFile(st_varent_network).extractall(gettempdir())
        self.st_varent = join(gettempdir(), "St_Varent")
        self.project = Project()
        self.project.open(self.st_varent)
        self.project.network.build_graphs()
        self.g = self.project.network.graphs["c"]  # type: Graph
        self.g.set_graph("distance")
        self.g.set_skimming("distance")

        self.r = PathResults()
        self.r.prepare(self.g)

        self.r.compute_path(387, 1067)
        self.project.close()
